from opynfield.config.defaults_settings import Defaults
from opynfield.config.user_input import UserInput
from opynfield.config.model_settings import ModelSpecification
from scipy.optimize import curve_fit
import pandas as pd
import numpy as np
import warnings


def fit_measure_by_time(
    group_individual_measure_df: pd.DataFrame, specs: ModelSpecification
) -> pd.DataFrame:
    """This function fits all the individuals in one group for a time vs y-measure model

    :param group_individual_measure_df: the necessary data
    :type group_individual_measure_df: pd.DataFrame
    :param specs: the model settings to use
    :type specs: ModelSpecification
    :return: a dataframe of the resultant parameter fits for each individual in the group
    :rtype: pd.DataFrame
    """
    # rows are individuals and columns are time points
    # measure to be modeled by time
    params_list = []
    for i in range(group_individual_measure_df.shape[0]):
        # for each individual
        y_raw = group_individual_measure_df.iloc[i].values[1:].astype(float)
        x_raw = np.arange(group_individual_measure_df.shape[1] - 1).astype(float)
        x_filter = x_raw[~np.isnan(y_raw)]
        y_filter = y_raw[~np.isnan(y_raw)]
        x = x_filter[~np.isnan(x_filter)]
        y = y_filter[~np.isnan(x_filter)]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Covariance")
            try:
                # noinspection PyTupleAssignmentBalance
                params, cv = curve_fit(
                    specs.model.model_function,
                    x,
                    y,
                    p0=specs.model.initial_params,
                    bounds=specs.model.bounds,
                    **{"maxfev": specs.model.max_eval},
                )
                params_list.append(pd.DataFrame(params))
            except ValueError:
                params = pd.DataFrame(
                    np.ones(
                        len(specs.model.initial_params),
                    )
                    * np.nan
                )
                params_list.append(params)
    params_df = pd.concat(params_list, axis=1).T
    # rows are individuals
    # columns are parameters
    return params_df


def fit_measure_by_cov(
    group_individual_measure_df: pd.DataFrame,
    group_individual_cov_df: pd.DataFrame,
    specs: ModelSpecification,
) -> pd.DataFrame:
    """This function fits all the individuals in one group for a coverage-measure vs y-measure model

    :param group_individual_measure_df: the necessary y measure data
    :type group_individual_measure_df: pd.DataFrame
    :param group_individual_cov_df: the necessary coverage measure data
    :type group_individual_cov_df: pd.DataFrame
    :param specs: the model settings to use
    :type specs: ModelSpecification
    :return: a dataframe of the resultant parameter fits for each individual in the group
    :rtype: pd.DataFrame
    """
    # rows are individuals and columns are time points
    # measure to be modeled by mode
    params_list = []
    for i in range(group_individual_measure_df.shape[0]):
        # for each individual
        y_raw = group_individual_measure_df.iloc[i].values[1:].astype(float)
        x_raw = group_individual_cov_df.iloc[i].values[1:].astype(float)
        x_filter = x_raw[~np.isnan(y_raw)]
        y_filter = y_raw[~np.isnan(y_raw)]
        x = x_filter[~np.isnan(x_filter)]
        y = y_filter[~np.isnan(x_filter)]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Covariance")
            try:
                # noinspection PyTupleAssignmentBalance
                params, cv = curve_fit(
                    specs.model.model_function,
                    x,
                    y,
                    p0=specs.model.initial_params,
                    bounds=specs.model.bounds,
                    **{"maxfev": specs.model.max_eval},
                )
                params_list.append(pd.DataFrame(params))
            except ValueError:
                params = pd.DataFrame(
                    np.ones(
                        len(specs.model.initial_params),
                    )
                    * np.nan
                )
                params_list.append(params)
    params_df = pd.concat(params_list, axis=1).T
    # rows are individuals
    # columns are parameters
    return params_df


def fit_by_time(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    time_averaged_measures: list[str],
    model_params: dict[str, dict[str, ModelSpecification]],
    group: str,
) -> dict[str, pd.DataFrame]:
    """This function coordinates fitting the models for all the individuals in a group for all the time vs y-measure relationships

    :param individual_measures_dfs: the necessary data in a dictionary indexed by group and y-measure
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param time_averaged_measures: the y-measures to be modeled against time
    :type time_averaged_measures: list[str]
    :param model_params: the model settings to use in a dictionary indexed by group and y-measure
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param group: the group to model
    :type group: str
    :return: the resultant parameter fits in a dictionary indexed by y-measure
    :rtype: dict[str, pd.DataFrame]
    """
    print(f"Fitting Models To Tracks From Group {group} by time")
    time_measures = {}
    for measure in time_averaged_measures:
        if measure != "r":
            specs = model_params["time"][measure]
            m_params = fit_measure_by_time(
                individual_measures_dfs[group][measure], specs
            )
            time_measures[measure] = m_params
    # dict key is measure
    # result is a df with rows as individuals and columns as parameters
    return time_measures


def fit_by_cov_measure(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    coverage_averaged_measures: list[str],
    model_params: dict[str, dict[str, ModelSpecification]],
    mode: str,
    group: str,
) -> dict[str, pd.DataFrame]:
    """This function coordinates fitting the models for all the individuals in a group for all the coverage-measure vs y-measure relationships

    :param individual_measures_dfs: the necessary data in a dictionary indexed by group and y-measure
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param coverage_averaged_measures: the y-measures to be modeled against coverage measures
    :type coverage_averaged_measures: list[str]
    :param model_params: the model settings to use in a dictionary indexed by group and y-measure
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param mode: the coverage measure to use as the x-measure
    :type mode: str
    :param group: the group to model
    :type group: str
    :return: the resultant parameter fits in a dictionary indexed by y-measure
    :rtype: dict[str, pd.DataFrame]
    """
    print(f"Fitting Models To Tracks From Group {group} by {mode}")
    cov_measures = {}
    for measure in coverage_averaged_measures:
        specs = model_params[mode][measure]
        m_params = fit_measure_by_cov(
            individual_measures_dfs[group][measure],
            individual_measures_dfs[group][mode],
            specs,
        )
        cov_measures[measure] = m_params
    # dict key is measure
    # result is a df with rows as individuals and columns as parameters
    return cov_measures


def fit_all(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    defaults: Defaults,
    model_params: dict[str, dict[str, ModelSpecification]],
) -> dict[str, dict[str, dict[str, pd.DataFrame]]]:
    """This function coordinates fitting the models for all the individuals in a all the groups for all the time and coverage-measure vs y-measure relationships

    :param individual_measures_dfs: the necessary data in a dictionary indexed by group and y-measure
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param model_params: the model settings to use in a dictionary indexed by group and y-measure
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :return: the resultant parameter fits in a dictionary indexed by group, x-measure, and y-measure
    :rtype: dict[str, dict[str, dict[str, pd.DataFrame]]]
    """
    # dict[group] is a dict 2
    # dict 2[measure] is a df with rows as individuals and columns as time points
    # returns a dict of [group] -> dict[x-axis] -> dict[measure] -> df
    fits = {}
    for group in individual_measures_dfs:
        fits_group = {
            "time": fit_by_time(
                individual_measures_dfs,
                defaults.time_averaged_measures,
                model_params,
                group,
            ),
            "coverage": fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "coverage",
                group,
            ),
            "pica": fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "pica",
                group,
            ),
            "pgca": fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "pgca",
                group,
            ),
            "percent_coverage": fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "percent_coverage",
                group,
            ),
        }
        fits[group] = fits_group
    return fits


def find_fit_bounds(
    fits: dict[str, dict[str, dict[str, pd.DataFrame]]], user_inputs: UserInput
) -> tuple[
    dict[str, dict[str, dict[str, pd.DataFrame]]],
    dict[str, dict[str, dict[str, pd.DataFrame]]],
    dict[str, dict[str, dict[str, pd.DataFrame]]],
]:
    """This function finds the bounds to use when re-fitting the individual bounds based on the distribution of the naive fits

    :param fits: the original (naive) parameter fits
    :type fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param user_inputs: user settings to use
    :type user_inputs: UserInput
    :return: the upper bounds, lower bounds, and p0s to use, each a dictionary indexed by group, x-measure, and y-measure
    :rtype: tuple[dict[str, dict[str, dict[str, pd.DataFrame]]], dict[str, dict[str, dict[str, pd.DataFrame]]], dict[str, dict[str, dict[str, pd.DataFrame]]]]
    """
    # dict of[group] -> dict of [x-axis] -> dict[measure] -> df
    upper_group_x_measure_dict = {}
    lower_group_x_measure_dict = {}
    mean_group_x_measure_dict = {}
    for group in fits:
        upper_x_measure_dict = {}
        lower_x_measure_dict = {}
        mean_x_measure_dict = {}
        for x_axis in fits[group]:
            upper_measure_dict = {}
            lower_measure_dict = {}
            mean_measure_dict = {}
            for measure in fits[group][x_axis]:
                df = fits[group][x_axis][measure]
                m = df.mean()
                s = df.std() * user_inputs.bound_level
                lower = m - s
                upper = m + s
                for i in range(len(upper)):
                    if np.sign(lower[i]) != np.sign(m[i]):
                        lower[i] = 0
                        # if they don't match signs -> move bound to 0
                        # if they do match signs -> keep the bound as is
                    if np.sign(upper[i]) != np.sign(m[i]):
                        upper[i] = 0
                        # if they don't match signs -> move bound to 0
                        # if they do match signs -> keep the bound as is
                upper_measure_dict[measure] = upper
                lower_measure_dict[measure] = lower
                mean_measure_dict[measure] = m
            upper_x_measure_dict[x_axis] = upper_measure_dict
            lower_x_measure_dict[x_axis] = lower_measure_dict
            mean_x_measure_dict[x_axis] = mean_measure_dict
        upper_group_x_measure_dict[group] = upper_x_measure_dict
        lower_group_x_measure_dict[group] = lower_x_measure_dict
        mean_group_x_measure_dict[group] = mean_x_measure_dict
    return (
        upper_group_x_measure_dict,
        lower_group_x_measure_dict,
        mean_group_x_measure_dict,
    )


def re_fit_measure_by_time(
    group_individual_measure_df: pd.DataFrame,
    specs: ModelSpecification,
    uppers: pd.DataFrame,
    lowers: pd.DataFrame,
    initials: pd.DataFrame,
) -> pd.DataFrame:
    """This function re-fits all the individuals in one group for a time vs y-measure model

    :param group_individual_measure_df: the necessary data
    :type group_individual_measure_df: pd.DataFrame
    :param specs: the model settings to use
    :type specs: ModelSpecification
    :param uppers: the upper bounds for model fits
    :type uppers: pd.DataFrame
    :param lowers: the lower bounds for model fits
    :type lowers: pd.DataFrame
    :param initials: the p0 for model fits
    :type initials: pd.DataFrame
    :return: a dataframe of the resultant parameter fits for each individual in the group
    :rtype: pd.DataFrame
    """
    # rows are individuals and columns are time points
    # measure to be modeled by time
    params_list = []
    for i in range(group_individual_measure_df.shape[0]):
        # for each individual
        y_raw = group_individual_measure_df.iloc[i].values[1:].astype(float)
        x_raw = np.arange(group_individual_measure_df.shape[1] - 1).astype(float)
        x_filter = x_raw[~np.isnan(y_raw)]
        y_filter = y_raw[~np.isnan(y_raw)]
        x = x_filter[~np.isnan(x_filter)]
        y = y_filter[~np.isnan(x_filter)]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Covariance")
            try:
                # noinspection PyTupleAssignmentBalance
                params, cv = curve_fit(
                    specs.model.model_function,
                    x,
                    y,
                    p0=initials.values,
                    bounds=(lowers.values, uppers.values),
                    **{"maxfev": specs.model.max_eval},
                )
                params_list.append(pd.DataFrame(params))
            except ValueError:
                params = pd.DataFrame(
                    np.ones(
                        len(specs.model.initial_params),
                    )
                    * np.nan
                )
                params_list.append(params)
    params_df = pd.concat(params_list, axis=1).T
    # rows are individuals
    # columns are parameters
    return params_df


def re_fit_by_time(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    time_averaged_measures: list[str],
    model_params: dict[str, dict[str, ModelSpecification]],
    upper: dict[str, dict[str, pd.DataFrame]],
    lower: dict[str, dict[str, pd.DataFrame]],
    initial: dict[str, dict[str, pd.DataFrame]],
    group: str,
) -> dict[str, pd.DataFrame]:
    """This function re-fits all the individuals in one group for all the time measures

    :param individual_measures_dfs: the necessary data indexed by group and y-measure
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param time_averaged_measures: which y-measures to model by time
    :type time_averaged_measures: list[str]
    :param model_params: the model settings to use indexed by group and y-measure
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param upper: the upper bounds for model fits indexed by group and y-measure
    :type upper: dict[str, dict[str, pd.DataFrame]]
    :param lower: the lower bounds for model fits indexed by group and y-measure
    :type lower: dict[str, dict[str, pd.DataFrame]]
    :param initial: the p0 for model fits indexed by group and y-measure
    :type initial: dict[str, dict[str, pd.DataFrame]]
    :param group: the group to run the models for
    :type group: str
    :return: a dataframe of the resultant parameter fits for each individual in the group
    :rtype: pd.DataFrame
    """
    print(f"Re-Fitting Models To Tracks From Group {group} by time")
    time_measures = {}
    for measure in time_averaged_measures:
        if measure != "r":
            specs = model_params["time"][measure]
            m_params = re_fit_measure_by_time(
                individual_measures_dfs[group][measure],
                specs,
                upper["time"][measure],
                lower["time"][measure],
                initial["time"][measure],
            )
            time_measures[measure] = m_params
    # dict key is measure
    # result is a dy with rows as individuals and columns as parameters
    return time_measures


def re_fit_measure_by_cov_measure(
    group_individual_measure_df: pd.DataFrame,
    group_individual_cov_df: pd.DataFrame,
    specs: ModelSpecification,
    uppers: pd.DataFrame,
    lowers: pd.DataFrame,
    initials: pd.DataFrame,
) -> pd.DataFrame:
    """This function re-fits all the individuals in one group for a coverage-measure vs y-measure model

    :param group_individual_measure_df: the necessary data for y measure
    :type group_individual_measure_df: pd.DataFrame
    :param group_individual_cov_df: the necessary data for coverage measure
    :type group_individual_cov_df: pd.DataFrame
    :param specs: the model settings to use
    :type specs: ModelSpecification
    :param uppers: the upper bounds for model fits
    :type uppers: pd.DataFrame
    :param lowers: the lower bounds for model fits
    :type lowers: pd.DataFrame
    :param initials: the p0 for model fits
    :type initials: pd.DataFrame
    :return: a dataframe of the resultant parameter fits for each individual in the group
    :rtype: pd.DataFrame
    """
    # rows are individuals and columns are time points
    # measure to be modeled by time
    params_list = []
    for i in range(group_individual_measure_df.shape[0]):
        # for each individual
        y_raw = group_individual_measure_df.iloc[i].values[1:].astype(float)
        x_raw = group_individual_cov_df.iloc[i].values[1:].astype(float)
        x_filter = x_raw[~np.isnan(y_raw)]
        y_filter = y_raw[~np.isnan(y_raw)]
        x = x_filter[~np.isnan(x_filter)]
        y = y_filter[~np.isnan(x_filter)]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="Covariance")
            try:
                # noinspection PyTupleAssignmentBalance
                params, cv = curve_fit(
                    specs.model.model_function,
                    x,
                    y,
                    p0=initials.values,
                    bounds=(lowers.values, uppers.values),
                    **{"maxfev": specs.model.max_eval},
                )
            except ValueError:
                params = (
                    np.ones(
                        len(specs.model.initial_params),
                    )
                    * np.nan
                )
        params_list.append(pd.DataFrame(params))
    params_df = pd.concat(params_list, axis=1).T
    # rows are individuals
    # columns are parameters
    return params_df


def re_fit_by_cov_measure(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    coverage_averaged_measures: list[str],
    model_params: dict[str, dict[str, ModelSpecification]],
    mode: str,
    upper: dict[str, dict[str, pd.DataFrame]],
    lower: dict[str, dict[str, pd.DataFrame]],
    initial: dict[str, dict[str, pd.DataFrame]],
    group: str,
) -> dict[str, pd.DataFrame]:
    """This function re-fits all the individuals in one group for all the coverage measures

    :param individual_measures_dfs: the necessary data indexed by group and y-measure
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param coverage_averaged_measures: which y-measures to model by coverage measure
    :type coverage_averaged_measures: list[str]
    :param model_params: the model settings to use indexed by group and y-measure
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param mode: which coverage measure to model with
    :type mode: str
    :param upper: the upper bounds for model fits indexed by group and y-measure
    :type upper: dict[str, dict[str, pd.DataFrame]]
    :param lower: the lower bounds for model fits indexed by group and y-measure
    :type lower: dict[str, dict[str, pd.DataFrame]]
    :param initial: the p0 for model fits indexed by group and y-measure
    :type initial: dict[str, dict[str, pd.DataFrame]]
    :param group: the group to run the models for
    :type group: str
    :return: a dataframe of the resultant parameter fits for each individual in the group
    :rtype: pd.DataFrame
    """
    print(f"Re-Fitting Models To Tracks From Group {group} by {mode}")
    cov_measures = {}
    for measure in coverage_averaged_measures:
        specs = model_params[mode][measure]
        m_params = re_fit_measure_by_cov_measure(
            individual_measures_dfs[group][measure],
            individual_measures_dfs[group][mode],
            specs,
            upper[mode][measure],
            lower[mode][measure],
            initial[mode][measure],
        )
        cov_measures[measure] = m_params
    # dict key is measure
    # result is a df with rows as individuals and columns as parameters
    return cov_measures


def re_fit_all(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    defaults: Defaults,
    model_params: dict[str, dict[str, ModelSpecification]],
    upper: dict[str, dict[str, dict[str, pd.DataFrame]]],
    lower: dict[str, dict[str, dict[str, pd.DataFrame]]],
    initial: dict[str, dict[str, dict[str, pd.DataFrame]]],
) -> dict[str, dict[str, dict[str, pd.DataFrame]]]:
    """This function re-fits all the individuals in one group for all the time and coverage measures

    :param individual_measures_dfs: the necessary data indexed by group and y-measure
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param model_params: the model settings to use indexed by group and y-measure
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param upper: the upper bounds for model fits indexed by group and y-measure and x-measure
    :type upper: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param lower: the lower bounds for model fits indexed by group and y-measure and x-measure
    :type lower: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param initial: the p0 for model fits indexed by group and y-measure and x-measure
    :type initial: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :return: a dataframe of the resultant parameter fits for each individual in the group indexed by the group y-measure and x-measure
    :rtype: dict[str, dict[str, dict[str, pd.DataFrame]]]
    """
    fits = {}
    for group in individual_measures_dfs:
        fits_group = {
            "time": re_fit_by_time(
                individual_measures_dfs,
                defaults.time_averaged_measures,
                model_params,
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
            "coverage": re_fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "coverage",
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
            "pica": re_fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "pica",
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
            "pgca": re_fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "pgca",
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
            "percent_coverage": re_fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "percent_coverage",
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
        }
        fits[group] = fits_group
    return fits



from opynfield.readin.track import Track
from tkinter import filedialog
import pandas as pd
import numpy as np
import datetime
from opynfield.readin import multi_tracker


def read_anymaze_center(
    groups_with_file_type: list[str],
    verbose: bool,
    arena_radius_cm: float,
    running_window_length: int,
    window_step_size: int,
    sample_freq: int,
    time_bin_size: int,
    trim,
    all_tracks: list[Track],
) -> list[Track]:
    """This function reads in all the tracks from the Anymaze tracker format that recorded the center position of the
    animal subject (rather than the head position). The function extracts the x, y, and t information and smooths,
    centers, and converts the units of the track.

    :param groups_with_file_type: which groups have tracks recorded in this type
    :type groups_with_file_type: list[str]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param arena_radius_cm: the radius of the arena in which the track was recorded, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type arena_radius_cm: float
    :param running_window_length: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type running_window_length: int
    :param window_step_size: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type window_step_size: int
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param trim: how many points are recorded before the animal enters the arena (maximum of all animals in this track type), sourced from :class:`opynfield.config.user_input.UserInput` object
    :type trim: int
    :param all_tracks: a list with all the Track objects from previously read-in datatypes
    :type all_tracks: list[Track]
    :return: a list of Track objects with a consistent format for x y and t tracking points
    :rtype: list[Track]
    """
    for anymaze_group in groups_with_file_type:
        if verbose:
            print(
                f"Running Anymaze Center Point Tracker Files For Group: {anymaze_group}"
            )
        title = f"Select .csv files for {anymaze_group}"
        # select .csv buri files for the groups
        data_files = filedialog.askopenfilenames(
            filetypes=[("CSV files", "*.csv")], title=title
        )
        for file_num in range(len(data_files)):
            if verbose:
                print(f"{anymaze_group}, File{file_num + 1} Out Of {len(data_files)}")
            # read in each file's data
            file_data = pd.read_csv(data_files[file_num], sep=",")#, lineterminator="\n")
            #file_data = file_data[:-1]
            track = Track(
                anymaze_group,
                file_data["Centre position X"].to_numpy(),
                file_data["Centre position Y"].to_numpy(),
                file_data["Time"].to_numpy(),
                "AnyMaze Center",
                [],
                False,
            )
            # put track coordinates through standardization procedures
            track.anymaze_center_numeric(verbose)
            track.t = convert_time_stamp(track.t, verbose)
            track.anymaze_center_running_line(
                running_window_length, window_step_size, verbose
            )
            track.anymaze_center_subsample(sample_freq, time_bin_size, verbose)
            track.anymaze_center_fill_missing(verbose)
            track_center = multi_tracker.calc_center(track.x, track.y, verbose, trim)
            track.anymaze_center_convert_to_center(track_center, verbose)
            track.anymaze_center_convert_units(arena_radius_cm, trim)
            track.standardized = True
            all_tracks.append(track)
    return all_tracks


def read_anymaze_head(
    groups_with_file_type: list[str],
    verbose: bool,
    arena_radius_cm: float,
    running_window_length: int,
    window_step_size: int,
    sample_freq: int,
    time_bin_size: int,
    trim,
    all_tracks: list[Track],
) -> list[Track]:
    """This function reads in all the tracks from the Anymaze tracker format that recorded the head position of the
    animal subject (rather than the body center point). The function extracts the x, y, and t information and smooths,
    centers, and converts the units of the track.

    :param groups_with_file_type: which groups have tracks recorded in this type
    :type groups_with_file_type: list[str]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param arena_radius_cm: the radius of the arena in which the track was recorded, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type arena_radius_cm: float
    :param running_window_length: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type running_window_length: int
    :param window_step_size: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type window_step_size: int
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param trim: how many points are recorded before the animal enters the arena (maximum of all animals in this track type), sourced from :class:`opynfield.config.user_input.UserInput` object
    :type trim: int
    :param all_tracks: a list with all the Track objects from previously read-in datatypes
    :type all_tracks: list[Track]
    :return: a Track object with a consistent format for x y and t tracking points
    :rtype: list[Track]
    """
    for anymaze_group in groups_with_file_type:
        if verbose:
            print(
                f"Running Anymaze Head Point Tracker Files For Group: {anymaze_group}"
            )
        title = f"Select .csv files for {anymaze_group}"
        # select .csv buri files for the groups
        data_files = filedialog.askopenfilenames(
            filetypes=[("CSV files", "*.csv")], title=title
        )
        for file_num in range(len(data_files)):
            if verbose:
                print(f"{anymaze_group}, File{file_num + 1} Out Of {len(data_files)}")
            # read in each file's data
            file_data = pd.read_csv(data_files[file_num], sep=",")#, lineterminator="\n")
            #file_data = file_data[:-1]
            track = Track(
                anymaze_group,
                file_data["Head position X"].to_numpy(),
                file_data["Head position Y"].to_numpy(),
                file_data["Time"].to_numpy(),
                "AnyMaze Head",
                [],
                False,
            )
            # put track coordinates through standardization procedures
            track.anymaze_head_numeric(verbose)
            track.t = convert_time_stamp(track.t, verbose)
            track.anymaze_head_running_line(
                running_window_length, window_step_size, verbose
            )
            track.anymaze_head_subsample(sample_freq, time_bin_size, verbose)
            track.anymaze_head_fill_missing(verbose)
            track_center = multi_tracker.calc_center(track.x, track.y, verbose, trim)
            track.anymaze_head_convert_to_center(track_center, verbose)
            track.anymaze_head_convert_units(arena_radius_cm, trim)
            track.standardized = True
            all_tracks.append(track)
    return all_tracks


def convert_time_stamp(time_stamp, verbose):
    """This function converts the time column from the form 00:00:00.00 (or similar) to the form 0.00
    Note: there seems to be some inconsistency with how the time stamp column is saved in the Anymaze format, please
    submit an error report if you encounter a new format that is not addressed in this function.

    :param time_stamp: the original time column
    :type time_stamp: np.ndarray
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: the re-formatted time column
    :rtype: np.ndarray
    """
    time_elapsed = np.zeros(len(time_stamp))  # initialize results
    d = datetime.date(2018, 11, 9)  # dummy date
    t_init = datetime.time.fromisoformat("0" + time_stamp[0])
    if verbose:
        print(f"Initial Time: {t_init}")
    t_dt_init = datetime.datetime.combine(d, t_init)  # time at start
    for i in range(len(time_stamp)):
        t_string = "0" + time_stamp[i]
        t_time = datetime.time.fromisoformat(t_string)
        t_datetime = datetime.datetime.combine(d, t_time)  # time at point
        t_delta = t_datetime - t_dt_init
        time_elapsed[i] = t_delta.total_seconds()  # time since start in seconds
    return time_elapsed



from dataclasses import dataclass


@dataclass
class PlotSettings:
    """This dataclass defines many settings that govern the design of the output plots

    Attributes:
        group_colors (dict[str, str]): what colors to use for each group in group comparison plots - must have one color per group
        marker_size (int): what size to plot the markers, defaults to 2
        marker_color (str): what color to plot the data (for single animals or single groups), defaults to 'b' (blue)
        individual_model_fit (bool): whether to plot the models on single animal plots, defaults to True
        fit_color (str): what color to plot the model (for single animals or single groups), defaults to 'k' (black)
        alpha (float): transparency to plot the model, defaults to 0.3
        group_error_bars (bool): whether to plot the error bars for group averages, defaults to True
        error_color (str): what color to plot the error bars for group averages, defaults to 'b' (blue)
        n_between_error (int): n to plot error bars on every nth data point, defaults to 1
        group_model_fit (bool): whether to plot the models on group plots, defaults to True
        equation (bool): whether to display the model equation on single animal or single group plots, defaults to True
        display_individual_figures (bool): whether to render the single animal plots, defaults to False
        save_individual_figures (bool): whether to save out the single animal plots, defaults to True
        display_solo_group_figures (bool): whether to render the group plots, defaults to False
        save_solo_group_figures (bool): whether to save out the single group plots, defaults to True
        save_combined_view_figures (bool): whether to save out single group plots that show component individuals, defaults to True
        fig_extension (str): what file format to save plots in, defaults to '.png'
        colormap_name (str): what color map to use for the trajectory plot time bar, defaults to 'gist_rainbow'
        edge_color (str): what color to plot the arena boundary for thr trajectory plot time bar, defaults to 'k' (black)
        error_width (float): what width to plot the error bars, defaults to 0.5
        save_group__comparison_figures (bool): whether to save out the the group comparison plots, defaults to True
    """
    # group_colors
    group_colors: dict[str, str]
    # size for the markers in scatter plots
    marker_size: int = 2
    # individual marker color
    marker_color: str = "b"
    # include model fit?
    individual_model_fit: bool = True
    # individual model color
    fit_color: str = "k"
    # individual model transparency
    alpha: float = 0.3
    # include error bars?
    group_error_bars: bool = True
    # color of group error_bars
    error_color: str = "b"
    # how many points to skip between errors shown
    n_between_error: int = 1
    # include model fits?
    group_model_fit: bool = True
    # include model equations?
    equation: bool = True
    # show the plots from individuals
    display_individual_figures: bool = False
    # save the plots from individuals
    save_individual_figures: bool = True
    # show the plots from solo groups
    display_solo_group_figures: bool = False
    # save the plots from solo groups
    save_solo_group_figures: bool = True
    # save the plots with individual and group combined view
    save_combined_view_figures: bool = True
    # figure format to save in
    fig_extension: str = ".png"
    # which colormap to use for track color bar
    colormap_name: str = "gist_rainbow"
    # what color to plot the arena edge
    edge_color: str = "k"
    # how thick the error bars should be
    error_width: float = 0.5
    # save the plots of group comparisons
    save_group__comparison_figures: bool = True



from dataclasses import dataclass


@dataclass
class Defaults:
    """This dataclass defines many default values that are used across the package.

    Attributes:
        node_size (float): the angle that (when combined with a radial bound) defines a bin for coverage calculations, defaults to 0.1
        save_group_csvs (bool): whether to save a separate .csv file of the component measures for each group, defaults to True
        save_all_group_csvs (bool): whether to save a combined .csv file of the component measures of each group, defaults to True
        save_group_model_csvs (bool): whether to save a separate .csv file of the model parameters for each group, defaults to True
        save_all_group_model_csvs (bool): whether to save a combined .csv file of the model parameters of each group, defaults to True
        n_points_coverage (int): the number of points to group together in an average for the coverage domain, defaults to 36
        n_points_pica (int): the number of points to group together in an average for the pica domain, defaults to 36
        n_points_pgca (int): the number of points to group together in an average for the pgca domain, defaults to 36
        n_bins_percent_coverage (int): the number of points to group together in an average for the percent coverage domain, defaults to 10
        time_averaged_measures (list[str]): which measures should be averaged in the time domain, defaults to ["r", "activity", "p_plus_plus", "p_plus_minus", "p_plus_zero", "p_zero_plus", "p_zero_zero", "coverage", "percent_coverage", "pica", "pgca", "p_plus_plus_given_plus", "p_plus_minus_given_plus", "p_plus_zero_given_plus", "p_zero_plus_given_zero", "p_zero_zero_given_zero", "p_plus_plus_given_any", "p_plus_minus_given_any", "p_plus_zero_given_any", "p_zero_plus_given_any", "p_zero_zero_given_any"]
        coverage_averaged_measures (list[str]): which measures should be averages in the coverage, percent coverage, pica, and pgca domains, defaults to ["activity", "p_plus_plus", "p_plus_minus", "p_plus_zero", "p_zero_plus", "p_zero_zero", "p_plus_plus_given_plus", "p_plus_minus_given_plus", "p_plus_zero_given_plus", "p_zero_plus_given_zero", "p_zero_zero_given_zero", "p_plus_plus_given_any", "p_plus_minus_given_any", "p_plus_zero_given_any", "p_zero_plus_given_any", "p_zero_zero_given_any"]
    """
    # what angle should we use to create the bins for coverage (degrees)
    node_size: float = 0.1
    # should we save out a csv for each group's component measures?
    save_group_csvs: bool = True
    # should we save out a csv with the measures from all the groups in it? (better for stats)
    # save_group_csvs must be true for save_all_group_csvs to be true
    save_all_group_csvs: bool = True
    # should we save a df of the parameters for individuals
    save_group_model_csvs: bool = True
    # should we save a df of the parameters for the individuals from all the groups in it (better for stats)
    save_all_group_model_csvs: bool = True
    # number of points to group together in an average
    n_points_coverage: int = 36
    # number of points to group together in an average
    n_points_pica: int = 36
    # number of points to group together in an average
    n_points_pgca: int = 36
    # number of bins to group together in an average
    n_bins_percent_coverage: int = 10
    # measures to time average (not all measures make sense to average (e.g. angular position))
    time_averaged_measures = [
        "r",
        "activity",
        "p_plus_plus",
        "p_plus_minus",
        "p_plus_zero",
        "p_zero_plus",
        "p_zero_zero",
        "coverage",
        "percent_coverage",
        "pica",
        "pgca",
        "p_plus_plus_given_plus",
        "p_plus_minus_given_plus",
        "p_plus_zero_given_plus",
        "p_zero_plus_given_zero",
        "p_zero_zero_given_zero",
        "p_plus_plus_given_any",
        "p_plus_minus_given_any",
        "p_plus_zero_given_any",
        "p_zero_plus_given_any",
        "p_zero_zero_given_any"
    ]
    coverage_averaged_measures = [
        "activity",
        "p_plus_plus",
        "p_plus_minus",
        "p_plus_zero",
        "p_zero_plus",
        "p_zero_zero",
        "p_plus_plus_given_plus",
        "p_plus_minus_given_plus",
        "p_plus_zero_given_plus",
        "p_zero_plus_given_zero",
        "p_zero_zero_given_zero",
        "p_plus_plus_given_any",
        "p_plus_minus_given_any",
        "p_plus_zero_given_any",
        "p_zero_plus_given_any",
        "p_zero_zero_given_any"
    ]

    def create_pairs(self):
        """Create a list of strings that indicate which measure combinations should be modeled and tested

        :return: list of strings that indicate xy pairs
        :rtype: list[str]
        """
        test_list = []
        for x in ["time", "coverage", "percent_coverage", "pica", "pgca"]:
            if x == "time":
                for y in self.time_averaged_measures:
                    if y != "r":
                        test_list.append(f"{x}_{y}_parameter_")
            else:
                for y in self.coverage_averaged_measures:
                    test_list.append(f"{x}_{y}_parameter_")
        return test_list



import sys
from dataclasses import fields
from collections import defaultdict

from opynfield.calculate_measures.standard_track import StandardTrack
from opynfield.config.cov_asymptote import CoverageAsymptote
from opynfield.config.defaults_settings import Defaults
from opynfield.config.model_settings import ModelSpecification
from opynfield.config.plot_settings import PlotSettings
from opynfield.config.user_input import UserInput


def summarize(
    tracks_by_group: defaultdict[str, list[StandardTrack]],
    test_cov_asymptote: CoverageAsymptote,
    user_defaults: Defaults,
    model_settings: dict[str, dict[str, ModelSpecification]],
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function summarizes information about the kinds of tracks read in after the read-in process is complete. It
    generates a text file of information about the tracks and parameters used on them

    :param tracks_by_group: the tracks to summarize
    :type tracks_by_group: defaultdict[str, list[StandardTrack]]
    :param test_cov_asymptote: the coverage asymptote information that was used
    :type test_cov_asymptote: CoverageAsymptote
    :param user_defaults: the default settings that were used
    :type user_defaults: Defaults
    :param model_settings: the model settings that were used
    :type model_settings: dict[str, dict[str, ModelSpecification]]
    :param plot_settings: the plot settings that were used
    :type plot_settings: PlotSettings
    :param user_config: the user inputs that were used
    :type user_config: UserInput
    """
    print("Tracks Run")
    print("__________")
    print("\n")
    print("N Per Group")
    for key, value in tracks_by_group.items():
        print(f"{key}: {len(value)}")
    print("\n")
    print("\n")
    print("Run Parameters")
    print("______________")
    print("\n")
    print("Coverage Asymptote Settings:")
    for field in fields(test_cov_asymptote):
        value = getattr(test_cov_asymptote, field.name)
        print(f"{field.name}: {value}")
    print("\n")
    print("Default Settings:")
    for field in fields(user_defaults):
        if field.name not in ["time_averaged_measures", "coverage_averaged_measures"]:
            value = getattr(user_defaults, field.name)
            print(f"{field.name}: {value}")
    print("\n")
    print("Model Settings:")
    for x_measure in model_settings:
        for y_measure in model_settings[x_measure]:
            print(f"{model_settings[x_measure][y_measure].axes}:")
            m = model_settings[x_measure][y_measure].model
            for field in fields(m):
                value = getattr(m, field.name)
                print(f"{field.name}: {value}")
    print("\n")
    print("Plot Settings:")
    for field in fields(plot_settings):
        value = getattr(plot_settings, field.name)
        print(f"{field.name}: {value}")
    print("\n")
    print("User Input Settings:")
    for field in fields(user_config):
        value = getattr(user_config, field.name)
        print(f"{field.name}: {value}")
    return


def summary_file(
    tracks_by_group: defaultdict[str, list[StandardTrack]],
    test_cov_asymptote: CoverageAsymptote,
    user_defaults: Defaults,
    model_settings: dict[str, dict[str, ModelSpecification]],
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function calls the summary information function and creates the text file output

    :param tracks_by_group: the tracks to summarize
    :type tracks_by_group: defaultdict[str, list[StandardTrack]]
    :param test_cov_asymptote: the coverage asymptote information that was used
    :type test_cov_asymptote: CoverageAsymptote
    :param user_defaults: the default settings that were used
    :type user_defaults: Defaults
    :param model_settings: the model settings that were used
    :type model_settings: dict[str, dict[str, ModelSpecification]]
    :param plot_settings: the plot settings that were used
    :type plot_settings: PlotSettings
    :param user_config: the user inputs that were used
    :type user_config: UserInput
    """
    print("Formatting Summary File")
    # set up the output file
    summary_file_path = user_config.result_path + "/summary.txt"
    orig_stdout = sys.stdout
    f = open(summary_file_path, "a")
    sys.stdout = f

    # do the summary stuff with print statements
    summarize(
        tracks_by_group,
        test_cov_asymptote,
        user_defaults,
        model_settings,
        plot_settings,
        user_config,
    )

    # close out the output file
    sys.stdout = orig_stdout
    f.close()
    return



from dataclasses import dataclass
import os


@dataclass
class UserInput:
    """This dataclass defines many user inputs that are needed to properly read in and analyze the data

    Attributes:
        groups_and_types (dict[str, list[str]): dictionary of group names to a list of filetypes that include tracks from that group
        groups_to_paths (dict[str, str]): if the group names include non-standard characters (e.g. '/'), how to display the group names without those nonstandard characters
        arena_radius_cm (dict[str, str]): radius of the arena the tracks were recorded in (in cm)
        sample_freq (int): the frame rate that tracking points were recorded with
        edge_dist_cm (float): how far into the arena is considered the edge region
        time_bin_size (int): how many seconds should be binned together in the aggregation
        inactivity_threshold (float): how small of a step should be considered body wobble rather than activity
        verbose (bool): whether to display progress updates
        result_path (str): path to folder where you want to store results
        running_window_length (int): smoothing function parameter set to match ethovision smoothing, defaults to 5
        window_step_size (int): smoothing function parameter set to match ethovision smoothing, defaults to 1
        trim (int): for recordings that start before the animal is in the arena, how many points to trim off the beginning so that the arena boundary is identified correctly, defaults to 0
        bound_level (float): how many standard deviations to use when bounding the parameters fits, defaults to 2
    """
    # key for group names, value for list of file_types that group was recorded with
    groups_and_types: dict[str, list[str]]
    # key for group names as found in data files, value for a variation of that name without forbidden characters
    groups_to_paths: dict[str, str]
    # radius of arena
    arena_radius_cm: float
    # frame rate for your tracking coordinates
    sample_freq: int
    # how far into the arena should we consider the 'edge'
    edge_dist_cm: float
    # how many seconds should we bin together to aggregate
    time_bin_size: int
    # how small of a step should we consider body wobble rather than activity
    inactivity_threshold: float
    # do you want to display progress updates
    verbose: bool
    # path to folder where you will store results
    result_path: str
    # smoothing function parameter set to match ethovision
    running_window_length: int = 5
    # smoothing function parameter set to match ethovision
    window_step_size: int = 1
    # how many points until animal is in the arena
    trim: int = 0
    # how many stds to bound parameter fits to
    bound_level: float = 2

    def set_edge_radius(self):
        """This method takes the arena radius and the edge distance to calculate the edge radiusd

        :return: the radius at which the arena edge region begins
        :rtype: float
        """
        edge_radius = self.arena_radius_cm - self.edge_dist_cm
        return edge_radius

    def change_running_window_length(self, new_window_length):
        """This method allows the running window length to be changed"""
        self.running_window_length = new_window_length

    def change_window_step_size(self, new_window_step_size):
        """This method allows the window step size to be changed"""
        self.window_step_size = new_window_step_size

    def change_trim(self, new_trim):
        """This method allows the trim parameter to be changed"""
        self.trim = new_trim

    def prep_directory(self):
        """This method creates the folder in which the results will be saved"""
        os.makedirs(self.result_path, exist_ok=True)



import os
from collections import defaultdict
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from opynfield.config.defaults_settings import Defaults
from opynfield.config.model_settings import ModelSpecification
from opynfield.config.plot_settings import PlotSettings
from opynfield.config.user_input import UserInput
from opynfield.calculate_measures.standard_track import StandardTrack
from itertools import chain, zip_longest


def generate_fig_title(
    path: str, i: int, x_measure: str, y_measure: str, model_fit: bool, extension: str
):
    """This function generates a figure title to save the figure from the component parts

    :param path: path to the plot folder
    :type path: str
    :param i: which individual in the group is being plotted
    :type i: int
    :param x_measure: the x-axis of the plot
    :type x_measure: str
    :param y_measure: the y-axis of the plot
    :type y_measure: str
    :param model_fit: whether the plot includes the model fit
    :type model_fit: bool
    :param extension: what file extension / format to save the plot in
    :type extension: str
    :return: the path that the plot should be saved in
    :rtype: str
    """
    path = path + f"individual_{i}_{x_measure}_vs_{y_measure}"
    if model_fit:
        path = path + "_with_model"
    path = path + extension
    return path


def plot_time_measure(
    group: str,
    i: int,
    measure: str,
    measure_data: pd.DataFrame,
    model_params: pd.DataFrame,
    model_info: ModelSpecification,
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function plots an individual's data for a time vs y-measure relationship

    :param group: the group name
    :type group: str
    :param i: the individual number within its group
    :type i: int
    :param measure: the y-measure
    :type measure: str
    :param measure_data: the y-measure data
    :type measure_data: pd.DataFrame
    :param model_params: the model parameters to use, indexed by group and y-measure
    :type model_params: pd.DataFrame
    :param model_info: the model settings to use
    :type model_info: ModelSpecification
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    # create figure and axes objects for the plot
    fig, ax = plt.subplots()
    # define the x and y to be plotted
    # y is the measure for the individual
    y_plot = measure_data.iloc[i][1:].values
    # x is time in seconds
    x_plot = np.arange(len(y_plot))
    # scatter plot
    ax.scatter(
        x_plot, y_plot, s=plot_settings.marker_size, c=plot_settings.marker_color
    )
    if plot_settings.individual_model_fit:
        # plot the model fit
        fit_params = model_params.iloc[i].values
        if ~np.isnan(fit_params[0]):  # if the fit was successful
            y_fit = model_info.model.model_function(x_plot, *fit_params)
            ax.plot(x_plot, y_fit, c=plot_settings.fit_color, alpha=plot_settings.alpha)
            if plot_settings.equation:
                equation_str = generate_individual_equation_str(
                    fit_params, model_info.model.display_parts
                )
                ax.set_title(f"model fit: {equation_str}")
    # add the axis labels
    ax.set_xlabel("time (s)")
    ax.set_ylabel(f"{measure}")
    # add the plot title
    fig.suptitle(f"{measure} by time individual {i} from group {group}")
    # set axis limits
    if measure not in ["activity", "percent_coverage", "pica", "pgca", "coverage"]:
        # ax.set_xlim()
        ax.set_ylim((-0.1, 1.1))
    if plot_settings.display_individual_figures:
        # show the figure
        fig.show()
    if plot_settings.save_individual_figures:
        # save the figure
        path = (
            user_config.result_path
            + "/Individuals/"
            + user_config.groups_to_paths[group]
            + "/Plots/by_time/"
        )
        os.makedirs(path, exist_ok=True)
        fig_path = generate_fig_title(
            path,
            i,
            "time",
            measure,
            plot_settings.individual_model_fit,
            plot_settings.fig_extension,
        )
        fig.savefig(fname=fig_path, bbox_inches="tight")
    plt.close(fig=fig)
    return


def truncate(n, decimals=0):
    """This function trims the exact parameter fit value to a set number of decimal places for display on a plot

    :param n: the parameter
    :type n: float
    :param decimals: the number of decimal places to trim to
    :type decimals: int
    :return: the trimmed parameter value
    :rtype: float
    """
    multiplier = 10**decimals
    return int(n * multiplier) / multiplier


def generate_individual_equation_str(params, display_parts):
    """This function generates the equation of the best fit model from the model display parts and the truncated
    parameter values

    :param params: the parameter fits
    :type params: list[float]
    :param display_parts: the display parts from the model information
    :type display_parts: list[str]
    :return: the full equation string
    :rtype: str
    """
    equation_parts = list(display_parts)
    string_params = [str(truncate(x, 4)) for x in params]
    parts = [
        x for x in chain(*zip_longest(equation_parts, string_params)) if x is not None
    ]
    display_equation = "".join(parts)
    return display_equation


def plot_cmeasure_measure(
    x_measure: str,
    group: str,
    i: int,
    measure: str,
    measure_data_x: pd.DataFrame,
    measure_data_y: pd.DataFrame,
    model_params: pd.DataFrame,
    model_info: ModelSpecification,
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function plots an individual's data for a coverage-measure vs y-measure relationship

    :param x_measure: the coverage measure name
    :type x_measure: str
    :param group: the group name
    :type group: str
    :param i: the individual number within its group
    :type i: int
    :param measure: the y-measure
    :type measure: str
    :param measure_data_x: the coverage-measure data
    :type measure_data_x: pd.DataFrame
    :param measure_data_y: the y-measure data
    :type measure_data_y: pd.DataFrame
    :param model_params: the model parameters to use, indexed by group and y-measure
    :type model_params: pd.DataFrame
    :param model_info: the model settings to use
    :type model_info: ModelSpecification
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    # create figure and axes objects for the plot
    fig, ax = plt.subplots()
    # define the x and y to be plotted
    # y is the measure for the individual
    y_plot = measure_data_x.iloc[i][1:].values
    # x is the coverage measure for the individual
    x_plot = measure_data_y.iloc[i][1:].values
    # scatter plot
    ax.scatter(
        x_plot, y_plot, s=plot_settings.marker_size, c=plot_settings.marker_color
    )
    if plot_settings.individual_model_fit:
        # plot the model fit
        fit_params = model_params.iloc[i].values
        if ~np.isnan(fit_params[0]):  # if the fit was successful
            y_fit = model_info.model.model_function(x_plot.astype(float), *fit_params)
            ax.plot(x_plot, y_fit, c=plot_settings.fit_color, alpha=plot_settings.alpha)
            if plot_settings.equation:
                equation_str = generate_individual_equation_str(
                    fit_params, model_info.model.display_parts
                )
                ax.set_title(f"model fit: {equation_str}")
    # add the axis labels
    ax.set_xlabel(f"{x_measure}")
    ax.set_ylabel(f"{measure}")
    # add the plot title
    fig.suptitle(f"{measure} by {x_measure} individual {i} from group {group}")
    # set axis limits
    if measure != "activity":
        # ax.set_xlim()
        ax.set_ylim((-0.1, 1.1))
    if plot_settings.display_individual_figures:
        # show the figure
        fig.show()
    if plot_settings.save_individual_figures:
        # save the figure
        path = (
            user_config.result_path
            + "/Individuals/"
            + user_config.groups_to_paths[group]
            + "/Plots/by_"
            + x_measure
            + "/"
        )
        os.makedirs(path, exist_ok=True)
        fig_path = generate_fig_title(
            path,
            i,
            x_measure,
            measure,
            plot_settings.individual_model_fit,
            plot_settings.fig_extension,
        )
        fig.savefig(fname=fig_path, bbox_inches="tight")
    plt.close(fig=fig)
    return


def plot_all_individuals_by_time(
    group: str,
    group_measures: dict[str, pd.DataFrame],
    group_model_params: dict[str, pd.DataFrame],
    model_info: dict[str, ModelSpecification],
    defaults: Defaults,
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function coordinates plotting all the individuals in a group for a time vs y-measure relationship

    :param group: the group name
    :type group: str
    :param group_measures: the y-measure data use, indexed by y-measure
    :type group_measures: dict[str, pd.DataFrame]
    :param group_model_params: the model parameters to use, indexed by y-measure
    :type group_model_params: dict[str, pd.DataFrame]
    :param model_info: the model settings to use, indexed by the y-measure
    :type model_info: dict[str, ModelSpecification]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    for measure in defaults.time_averaged_measures:
        if measure != "r":
            for i in range(group_measures[measure].shape[0]):
                plot_time_measure(
                    group,
                    i,
                    measure,
                    group_measures[measure],
                    group_model_params[measure].reset_index().drop(columns="index"),
                    model_info[measure],
                    plot_settings,
                    user_config,
                )
    return


def plot_all_individuals_by_cmeasure(
    x_measure: str,
    group: str,
    group_measures: dict[str, pd.DataFrame],
    group_model_params: dict[str, pd.DataFrame],
    model_info: dict[str, ModelSpecification],
    defaults: Defaults,
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function coordinates plotting all the individuals in a group for a coverage-measure vs y-measure relationship

    :param x_measure: the coverage measure name
    :type x_measure: str
    :param group: the group name
    :type group: str
    :param group_measures: the y-measure data use, indexed by y-measure
    :type group_measures: dict[str, pd.DataFrame]
    :param group_model_params: the model parameters to use, indexed by y-measure
    :type group_model_params: dict[str, pd.DataFrame]
    :param model_info: the model settings to use, indexed by the y-measure
    :type model_info: dict[str, ModelSpecification]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    for measure in defaults.coverage_averaged_measures:
        for i in range(group_measures[measure].shape[0]):
            plot_cmeasure_measure(
                x_measure,
                group,
                i,
                measure,
                group_measures[measure],
                group_measures[x_measure],
                group_model_params[measure].reset_index().drop(columns="index"),
                model_info[measure],
                plot_settings,
                user_config,
            )

    pass


def plot_all_individuals(
    measures: dict[str, dict[str, pd.DataFrame]],
    model_params: dict[str, dict[str, dict[str, pd.DataFrame]]],
    model_info: dict[str, dict[str, ModelSpecification]],
    defaults: Defaults,
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function coordinates the plotting of all the x-measure vs y-measure relationships from all individuals from all groups

    :param measures: the y-measure data indexed by group and y-measure name
    :type measures: dict[str, dict[str, pd.DataFrame]]
    :param model_params: the model fit parameters indexed by group, x-measure, and y-measure
    :type model_params: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param model_info: the model settings to use, indexed by x-measure and y-measure
    :type model_info: dict[str, dict[str, ModelSpecification]]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    for group in measures:
        # plot all individuals by time
        print(f"Plotting Individuals From Group {group} by time")
        plot_all_individuals_by_time(
            group,
            measures[group],
            model_params[group]["time"],
            model_info["time"],
            defaults,
            plot_settings,
            user_config,
        )
        # plot all individuals by coverage measure
        for x_measure in ["coverage", "pica", "pgca", "percent_coverage"]:
            print(f"Plotting Individuals From Group {group} by {x_measure}")
            plot_all_individuals_by_cmeasure(
                x_measure,
                group,
                measures[group],
                model_params[group][x_measure],
                model_info[x_measure],
                defaults,
                plot_settings,
                user_config,
            )
    return


def plot_individual_trace(
    track: StandardTrack,
    i: int,
    group: str,
    plot_settings: PlotSettings,
    user_input: UserInput,
):
    """This function plots the trace (tracking trajectory) of an individual

    :param track: the track whose trace you are plotting
    :type track: StandardTrack
    :param i: the individual number within the tracks group
    :type i: int
    :param group: the group name to which the track belongs
    :type group: str
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_input: the user inputs to use
    :type user_input: UserInput
    """
    # create figure and axes objects for the plot
    fig, ax = plt.subplots()
    # define the x and y to be plotted
    x_starts = track.x[:-1]
    y_starts = track.y[:-1]
    x_stops = track.x[1:]
    y_stops = track.y[1:]
    t_scaled = np.arange(len(track.x))
    cmap = matplotlib.cm.get_cmap(plot_settings.colormap_name)
    # plot each step segment and color it by the time
    for j in range(len(x_starts)):
        ax.plot(
            [x_starts[j], x_stops[j]],
            [y_starts[j], y_stops[j]],
            c=cmap(t_scaled[j]),
            alpha=plot_settings.alpha,
        )
    # plot the arena
    angle = np.linspace(0, 2 * np.pi, 150)
    x = user_input.arena_radius_cm * np.cos(angle)
    y = user_input.arena_radius_cm * np.sin(angle)
    ax.plot(x, y, c=plot_settings.edge_color)
    ax.set_xlabel("X Coordinate (cm)")
    ax.set_ylabel("Y Coordinate (cm)")
    fig.suptitle(f"Individual {i} From Group {group} Track Trace")
    norm = matplotlib.colors.Normalize(vmin=t_scaled[0], vmax=t_scaled[-1])
    fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap), ax=plt.gca())
    if plot_settings.display_individual_figures:
        # show the figure
        fig.show()
    if plot_settings.save_individual_figures:
        # save the figure
        path = (
            user_input.result_path
            + "/Individuals/"
            + user_input.groups_to_paths[group]
            + "/Plots/traces/"
        )
        os.makedirs(path, exist_ok=True)
        fig_path = path + "individual_" + str(i) + plot_settings.fig_extension
        fig.savefig(fname=fig_path, bbox_inches="tight")
    plt.close(fig=fig)
    return


def plot_traces(
    tracks_by_groups: defaultdict[str, list[StandardTrack]],
    plot_settings: PlotSettings,
    user_input: UserInput,
):
    """This function coordinated plotting the traces of all individual tracks in all groups

    :param tracks_by_groups: a dictionary of all the individuals to plot, indexed by the group to which they belong
    :type tracks_by_groups: defaultdict[str, list[StandardTrack]]
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_input: the user inputs to use
    :type user_input: UserInput
    """
    for group in tracks_by_groups:
        for i, track in enumerate(tracks_by_groups[group]):
            plot_individual_trace(track, i, group, plot_settings, user_input)
    return



import pandas as pd
import numpy as np
import warnings
from scipy.optimize import curve_fit
from opynfield.config.defaults_settings import Defaults
from opynfield.config.model_settings import ModelSpecification


def group_fit_measure_by_time(
    group_individual_measure_df: pd.DataFrame,
    specs: ModelSpecification,
    uppers: pd.DataFrame,
    lowers: pd.DataFrame,
    initials: pd.DataFrame,
) -> pd.DataFrame:
    """This function fits the model for a single time vs y-measure relationship

    :param group_individual_measure_df: dataframe that includes the needed data
    :type group_individual_measure_df: pd.DataFrame
    :param specs: the model settings for this relationship
    :type specs: ModelSpecification
    :param uppers: the upper bounds for this relationship
    :type uppers: pd.DataFrame
    :param lowers: the lower bounds for this relationship
    :type lowers: pd.DataFrame
    :param initials: the p0 for this relationship
    :type initials: pd.DataFrame
    :return: the parameter fits for this relationship
    :rtype: pd.DataFrame
    """
    # rows are individuals and columns are time points
    # measure to be modeled by time
    # need to combine the coordinates from all the individuals to get 'group' coordinates
    y_raw = (
        group_individual_measure_df.drop(columns=["Group"])
        .values.flatten()
        .astype(float)
    )
    x_raw = np.tile(
        np.arange(group_individual_measure_df.shape[1] - 1).astype(float),
        group_individual_measure_df.shape[0],
    )
    x_filter = x_raw[~np.isnan(y_raw)]
    y_filter = y_raw[~np.isnan(y_raw)]
    x = x_filter[~np.isnan(x_filter)]
    y = y_filter[~np.isnan(x_filter)]
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Covariance")
        # noinspection PyTupleAssignmentBalance
        try:
        	params, cv = curve_fit(
           	 specs.model.model_function,
            	x,
            	y,
            	p0=initials.values,
            	bounds=(lowers.values, uppers.values),
            	**{"maxfev": specs.model.max_eval},
        	)
        except ValueError:
        	params, cv = curve_fit(
           	 specs.model.model_function,
            	x,
            	y,
            	p0=initials.values,
            	**{"maxfev": specs.model.max_eval},
        	)
    return pd.DataFrame(params)


def group_fit_measure_by_cov_measure(
    group_individual_measure_df: pd.DataFrame,
    group_individual_cov_df: pd.DataFrame,
    specs: ModelSpecification,
    uppers: pd.DataFrame,
    lowers: pd.DataFrame,
    initials: pd.DataFrame,
) -> pd.DataFrame:
    """This function fits the model for a single coverage-measure vs y-measure relationship

    :param group_individual_measure_df: dataframe that includes the needed y data
    :type group_individual_measure_df: pd.DataFrame
    :param group_individual_cov_df: the dataframe that includes the needed coverage data
    :type group_individual_cov_df: pd.DataFrame
    :param specs: the model settings for this relationship
    :type specs: ModelSpecification
    :param uppers: the upper bounds for this relationship
    :type uppers: pd.DataFrame
    :param lowers: the lower bounds for this relationship
    :type lowers: pd.DataFrame
    :param initials: the p0 for this relationship
    :type initials: pd.DataFrame
    :return: the parameter fits for this relationship
    :rtype: pd.DataFrame
    """
    # rows are individuals and columns are time points
    # measure to be modeled by time
    # need to combine the coordinates from all the individuals to get 'group' coordinates
    y_raw = (
        group_individual_measure_df.drop(columns=["Group"])
        .values.flatten()
        .astype(float)
    )
    x_raw = (
        group_individual_cov_df.drop(columns=["Group"]).values.flatten().astype(float)
    )
    x_filter = x_raw[~np.isnan(y_raw)]
    y_filter = y_raw[~np.isnan(y_raw)]
    x = x_filter[~np.isnan(x_filter)]
    y = y_filter[~np.isnan(x_filter)]
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Covariance")
        # noinspection PyTupleAssignmentBalance
        try:
        	params, cv = curve_fit(
            	specs.model.model_function,
            	x,
            	y,
            	p0=initials.values,
            	bounds=(lowers.values, uppers.values),
            	**{"maxfev": specs.model.max_eval},
        	)
        except ValueError:
        	params, cv = curve_fit(
            	specs.model.model_function,
            	x,
            	y,
            	p0=initials.values,
            	**{"maxfev": specs.model.max_eval},
        	)
    return pd.DataFrame(params)


def group_fit_by_time(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    time_averaged_measures: list[str],
    model_params: dict[str, dict[str, ModelSpecification]],
    upper: dict[str, dict[str, pd.DataFrame]],
    lower: dict[str, dict[str, pd.DataFrame]],
    initial: dict[str, dict[str, pd.DataFrame]],
    group: str,
) -> dict[str, pd.DataFrame]:
    """This function coordinates fitting all the models for time vs y-measures within a group

    :param individual_measures_dfs: dictionary that contains y-measure data indexed by groups and measure name
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param time_averaged_measures: list of all y measures to model
    :type time_averaged_measures: list[str]
    :param model_params: dictionary that contains model settings indexed by groups and measure name
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param upper: dictionary that contains upper parameter bounds indexed by groups and measure name
    :type upper: dict[str, dict[str, pd.DataFrame]]
    :param lower: dictionary that contains lower parameter bounds indexed by groups and measure name
    :type lower: dict[str, dict[str, pd.DataFrame]]
    :param initial: dictionary that contains p0 indexed by groups and measure name
    :type initial: dict[str, dict[str, pd.DataFrame]]
    :param group: what group to model
    :type group: str
    :return: a dictionary of the parameter fits by measure name for this group
    :rtype: dict[str, pd.DataFrame]
    """
    print(f"Fitting Models To Entire Group {group} by time")
    time_measures = {}
    for measure in time_averaged_measures:
        if measure != "r":
            specs = model_params["time"][measure]
            m_params = group_fit_measure_by_time(
                individual_measures_dfs[group][measure],
                specs,
                upper["time"][measure],
                lower["time"][measure],
                initial["time"][measure],
            )
            time_measures[measure] = m_params
    # dict key is measure
    # result is a df with columns as parameters (only 1 row for the group's fit)
    return time_measures


def group_fit_by_cov_measure(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    coverage_averaged_measures: list[str],
    model_params: dict[str, dict[str, ModelSpecification]],
    mode: str,
    upper: dict[str, dict[str, pd.DataFrame]],
    lower: dict[str, dict[str, pd.DataFrame]],
    initial: dict[str, dict[str, pd.DataFrame]],
    group: str,
) -> dict[str, pd.DataFrame]:
    """This function coordinates fitting all the models for time vs y-measures within a group

    :param individual_measures_dfs: dictionary that contains y-measure data indexed by groups and measure name
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param coverage_averaged_measures: list of all y measures to model
    :type coverage_averaged_measures: list[str]
    :param model_params: dictionary that contains model settings indexed by groups and measure name
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param mode: which coverage measure to use as the x-measure
    :type mode: str
    :param upper: dictionary that contains upper parameter bounds indexed by groups and measure name
    :type upper: dict[str, dict[str, pd.DataFrame]]
    :param lower: dictionary that contains lower parameter bounds indexed by groups and measure name
    :type lower: dict[str, dict[str, pd.DataFrame]]
    :param initial: dictionary that contains p0 indexed by groups and measure name
    :type initial: dict[str, dict[str, pd.DataFrame]]
    :param group: what group to model
    :type group: str
    :return: a dictionary of the parameter fits by measure name for this group
    :rtype: dict[str, pd.DataFrame]
    """
    print(f"Fitting Models To Entire Group {group} by {mode}")
    cov_measures = {}
    for measure in coverage_averaged_measures:
        specs = model_params[mode][measure]
        m_params = group_fit_measure_by_cov_measure(
            individual_measures_dfs[group][measure],
            individual_measures_dfs[group][mode],
            specs,
            upper[mode][measure],
            lower[mode][measure],
            initial[mode][measure],
        )
        cov_measures[measure] = m_params
    # dict key is measure
    # result is a df with columns as parameters (only 1 row for the group's fit)
    return cov_measures


def group_fit_all(
    individual_measures_dfs: dict[str, dict[str, pd.DataFrame]],
    defaults: Defaults,
    model_params: dict[str, dict[str, ModelSpecification]],
    upper: dict[str, dict[str, dict[str, pd.DataFrame]]],
    lower: dict[str, dict[str, dict[str, pd.DataFrame]]],
    initial: dict[str, dict[str, dict[str, pd.DataFrame]]],
) -> dict[str, dict[str, dict[str, pd.DataFrame]]]:
    """This function coordinates fitting all the models for time vs y-measures within a group

    :param individual_measures_dfs: dictionary that contains y-measure data indexed by groups and measure name
    :type individual_measures_dfs: dict[str, dict[str, pd.DataFrame]]
    :param defaults: what default settings to use
    :type defaults: Defaults
    :param model_params: dictionary that contains model settings indexed by groups and measure name
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param upper: dictionary that contains upper parameter bounds indexed by groups and measure name
    :type upper: dict[str, dict[str, pd.DataFrame]]
    :param lower: dictionary that contains lower parameter bounds indexed by groups and measure name
    :type lower: dict[str, dict[str, pd.DataFrame]]
    :param initial: dictionary that contains p0 indexed by groups and measure name
    :type initial: dict[str, dict[str, pd.DataFrame]]
    :return: a dictionary of the parameter fits by measure name for this group
    :rtype: dict[str, pd.DataFrame]
    """
    fits = {}
    for group in individual_measures_dfs:
        fits_group = {
            "time": group_fit_by_time(
                individual_measures_dfs,
                defaults.time_averaged_measures,
                model_params,
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
            "coverage": group_fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "coverage",
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
            "pica": group_fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "pica",
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
            "pgca": group_fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "pgca",
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
            "percent_coverage": group_fit_by_cov_measure(
                individual_measures_dfs,
                defaults.coverage_averaged_measures,
                model_params,
                "percent_coverage",
                upper[group],
                lower[group],
                initial[group],
                group,
            ),
        }
        fits[group] = fits_group
    return fits



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from opynfield.config.model_settings import ModelSpecification
from opynfield.config.defaults_settings import Defaults
from opynfield.config.plot_settings import PlotSettings
from opynfield.config.user_input import UserInput


def generate_fig_title(
    path: str,
    x_measure: str,
    y_measure: str,
    model_fit: bool,
    error: bool,
    extension: str,
) -> str:
    """This function generates a figure title to save the figure in from the component parts

    :param path: path to the plot folder
    :type path: str
    :param x_measure: the x-axis of the plot
    :type x_measure: str
    :param y_measure: the y-axis of the plot
    :type y_measure: str
    :param model_fit: whether the plot includes a model fit
    :type model_fit: bool
    :param error: whether the plot includes error bars
    :type error: bool
    :param extension: what file extension / format to save the plot in
    :type extension: str
    :return: the path to save the plot in
    :rtype: str
    """
    path = path + f"{x_measure}_vs_{y_measure}"
    if model_fit:
        path = path + "_with_model"
    if model_fit and error:
        path = path + "_and"
    if error:
        path = path + "_with_error"
    path = path + extension
    return path


def plot_time_comparison(
    x_measure: str,
    y_measure: str,
    time_averages: dict[str, dict[str, pd.DataFrame]],
    fits: dict[str, dict[str, dict[str, pd.DataFrame]]],
    specs: ModelSpecification,
    plot_settings: PlotSettings,
    user_inputs: UserInput,
):
    """This function plots all the group averages against each other for a time vs y-measure relationship

    :param x_measure: the x-measure (time)
    :type x_measure: str
    :param y_measure: the y-measure
    :type y_measure: str
    :param time_averages: the averages to use to plot, indexed by group and y-measure
    :type time_averages: dict[str, dict[str, pd.DataFrame]]
    :param fits: the model parameters to use, indexed by group and y-measure
    :type fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param specs: the model settings to use
    :type specs: ModelSpecification
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_inputs: the user inputs to use
    :type user_inputs: UserInput
    """
    # time_averages -> group -> y_measure -> df with averages and sems
    # fits -> group -> x_measure -> y_measure -> df with parameters

    # create figure and axes objects for the plot
    fig, ax = plt.subplots()

    # loop through groups for each step so no layers work out right

    # first do error bars if needed
    if plot_settings.group_error_bars:
        for group in time_averages:
            # define that group's x and y values
            y_plot = time_averages[group][y_measure].iloc[0][2:]
            y_error = time_averages[group][y_measure].iloc[1][2:]
            x_plot = np.arange(len(y_plot))
            ax.errorbar(
                x_plot,
                y_plot,
                yerr=y_error,
                xerr=None,
                fmt="none",
                ecolor=plot_settings.group_colors[group],
                errorevery=plot_settings.n_between_error,
                elinewidth=plot_settings.error_width,
                alpha=plot_settings.alpha,
            )
    # then do the scatter plots for all groups:
    for group in time_averages:
        y_plot = time_averages[group][y_measure].iloc[0][2:]
        x_plot = np.arange(len(y_plot))
        ax.scatter(
            x_plot,
            y_plot,
            s=plot_settings.marker_size,
            c=plot_settings.group_colors[group],
            label=group,
        )

    # then do the model fits if needed:
    if plot_settings.group_model_fit:
        for group in time_averages:
            y_plot = time_averages[group][y_measure].iloc[0][2:]
            x_plot = np.arange(len(y_plot))
            params = fits[group][x_measure][y_measure].values
            y_fit = specs.model.model_function(x_plot, *params)
            ax.plot(
                x_plot,
                y_fit,
                c=plot_settings.group_colors[group],
                alpha=plot_settings.alpha,
            )

    # add the axis labels
    ax.set_xlabel("time (s)")
    ax.set_ylabel(f"{y_measure}")

    # add the plot title
    fig.suptitle(f"{y_measure} by time")

    # add the legend of groups
    ax.legend()

    # set axis limits
    if y_measure not in ["activity", "percent_coverage", "pica", "pgca", "coverage"]:
        # ax.set_xlim()
        ax.set_ylim((-0.1, 1.1))

    if plot_settings.display_solo_group_figures:
        # show the figure
        fig.show()

    if plot_settings.save_group__comparison_figures:
        # save the figure
        path = user_inputs.result_path + "/GroupComparisonPlots/by_time/"
        os.makedirs(path, exist_ok=True)
        fig_path = generate_fig_title(
            path,
            "time",
            y_measure,
            plot_settings.group_model_fit,
            plot_settings.group_error_bars,
            plot_settings.fig_extension,
        )
        fig.savefig(fname=fig_path, bbox_inches="tight")
    plt.close(fig=fig)
    return


def plot_cmeasure_comparison(
    x_measure: str,
    y_measure: str,
    cmeasure_averages: dict[str, pd.DataFrame],
    group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]],
    specs: ModelSpecification,
    plot_settings: PlotSettings,
    user_input: UserInput,
):
    """This function plots all the group averages against each other for a coverage-measure vs y-measure relationship

    :param x_measure: the coverage-measure
    :type x_measure: str
    :param y_measure: the y-measure
    :type y_measure: str
    :param cmeasure_averages: the averages to use to plot, indexed by group and y-measure
    :type cmeasure_averages: dict[str, dict[str, pd.DataFrame]]
    :param group_fits: the model parameters to use, indexed by group and y-measure
    :type group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param specs: the model settings to use
    :type specs: ModelSpecification
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_input: the user inputs to use
    :type user_input: UserInput
    """
    # cmeasure_averages -> group -> df with averages and sems by name
    # fits -> group -> x_measure -> y_measure -> df with parameters

    # create figure and axes objects for the plot
    fig, ax = plt.subplots()

    # loop through groups for each step so no layers work out right
    # first do error bars if needed
    if plot_settings.group_error_bars:
        for group in cmeasure_averages:
            # define that group's x and y values
            y_plot = cmeasure_averages[group][f"{y_measure} mean"]
            y_error = cmeasure_averages[group][f"{y_measure} sem"]
            x_plot = cmeasure_averages[group][f"{x_measure} mean"]
            x_error = cmeasure_averages[group][f"{x_measure} sem"]
            ax.errorbar(
                x_plot,
                y_plot,
                yerr=y_error,
                xerr=x_error,
                fmt="none",
                ecolor=plot_settings.group_colors[group],
                errorevery=plot_settings.n_between_error,
                elinewidth=plot_settings.error_width,
                alpha=plot_settings.alpha,
            )
    # then do the scatter plots for all groups:
    for group in cmeasure_averages:
        y_plot = cmeasure_averages[group][f"{y_measure} mean"]
        x_plot = cmeasure_averages[group][f"{x_measure} mean"]
        ax.scatter(
            x_plot,
            y_plot,
            s=plot_settings.marker_size,
            c=plot_settings.group_colors[group],
            label=group,
        )

    # then do the model fits if needed:
    if plot_settings.group_model_fit:
        for group in cmeasure_averages:
            x_plot = cmeasure_averages[group][f"{x_measure} mean"]
            params = group_fits[group][x_measure][y_measure].values
            y_fit = specs.model.model_function(x_plot, *params)
            ax.plot(
                x_plot,
                y_fit,
                c=plot_settings.group_colors[group],
                alpha=plot_settings.alpha,
            )

    # add the axis labels
    ax.set_xlabel(f"f{x_measure}")
    ax.set_ylabel(f"{y_measure}")

    # add the plot title
    fig.suptitle(f"{y_measure} by {x_measure}")

    # add the legend of groups
    ax.legend()

    # set axis limits
    if y_measure != "activity":
        # ax.set_xlim()
        ax.set_ylim((-0.1, 1.1))

    if plot_settings.display_solo_group_figures:
        # show the figure
        fig.show()

    if plot_settings.save_group__comparison_figures:
        # save the figure
        path = user_input.result_path + "/GroupComparisonPlots/by_" + x_measure + "/"
        os.makedirs(path, exist_ok=True)
        fig_path = generate_fig_title(
            path,
            x_measure,
            y_measure,
            plot_settings.group_model_fit,
            plot_settings.group_error_bars,
            plot_settings.fig_extension,
        )
        fig.savefig(fname=fig_path, bbox_inches="tight")
    plt.close(fig=fig)
    return


def plot_all_group_comparisons(
    group_averages: dict[str, dict],
    group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]],
    model_params: dict[str, dict[str, ModelSpecification]],
    test_defaults: Defaults,
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function coordinates the plotting of all the x-measure (both time and coverage-measure) vs y-measure relationships

    :param group_averages: the group average data to use, indexed by group and then x measure
    :type group_averages: dict[str, dict]
    :param group_fits: the model parameter fits to use, indexed by group, x-measure, and y-measure
    :type group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param model_params: the model settings to use, indexed by x measure and y measure
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param test_defaults: the default settings to use
    :type test_defaults: Defaults
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    # do time plots
    print("Plotting Group Comparisons by time")
    for y_measure in test_defaults.time_averaged_measures:
        if y_measure != "r":
            plot_time_comparison(
                "time",
                y_measure,
                group_averages["time"],
                group_fits,
                model_params["time"][y_measure],
                plot_settings,
                user_config,
            )

    # do cmeasure plots
    for x_measure in ["coverage", "pica", "pgca", "percent_coverage"]:
        print(f"Plotting Group Comparisons by {x_measure}")
        for y_measure in test_defaults.coverage_averaged_measures:
            # plot_cmeasure_comparison()
            plot_cmeasure_comparison(
                x_measure,
                y_measure,
                group_averages[x_measure],
                group_fits,
                model_params[x_measure][y_measure],
                plot_settings,
                user_config,
            )
    return



from opynfield.readin import buridian_tracker
from opynfield.readin import etho_tracker
from opynfield.readin import anymaze_tracker

from opynfield.readin.track import Track


def read_track_types(
    file_type: str,
    groups_with_file_type: list[str],
    verbose: bool,
    arena_radius_cm: float,
    running_window_length: int,
    window_step_size: int,
    sample_freq: int,
    time_bin_size: int,
    trim,
    all_tracks: list[Track],
) -> list[Track]:
    """This function coordinates reading in all the data from each track type

    :param file_type: the file type to read in
    :type file_type: str
    :param groups_with_file_type: which groups have tracks recorded in that file type
    :type groups_with_file_type: list[str]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param arena_radius_cm: the radius of the arena in which the track was recorded (only required for buridian and anymaze trackers), sourced from :class:`opynfield.config.user_input.UserInput` object
    :type arena_radius_cm: float
    :param running_window_length: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type running_window_length: int
    :param window_step_size: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type window_step_size: int
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param trim: how many points are recorded before the animal enters the arena (maximum of all animals in this track type), sourced from :class:`opynfield.config.user_input.UserInput` object
    :type trim: int
    :param all_tracks: a list with all the Track objects from previously read-in datatypes
    :type all_tracks: list[Track]
    :return: a list of Track objects with a consistent format for x y and t tracking points
    :rtype: list[Track]
    """
    if file_type == "Buridian Tracker":
        all_tracks = buridian_tracker.read_buridian(
            groups_with_file_type,
            verbose,
            arena_radius_cm,
            running_window_length,
            window_step_size,
            sample_freq,
            time_bin_size,
            all_tracks,
        )
    if file_type == "Ethovision Excel Version 1":
        all_tracks = etho_tracker.read_etho_v1(
            groups_with_file_type, verbose, sample_freq, time_bin_size, all_tracks
        )
    if file_type == "Ethovision Excel Version 2":
        all_tracks = etho_tracker.read_etho_v2(
            groups_with_file_type, verbose, sample_freq, time_bin_size, all_tracks
        )
    if file_type == "Ethovision Text":
        all_tracks = etho_tracker.read_etho_txt(
            groups_with_file_type, verbose, sample_freq, time_bin_size, all_tracks
        )
    if file_type == "Ethovision Through MATLAB":
        all_tracks = etho_tracker.read_etho_ml(
            groups_with_file_type, verbose, sample_freq, time_bin_size, all_tracks
        )
    if file_type == "AnyMaze Center":
        all_tracks = anymaze_tracker.read_anymaze_center(
            groups_with_file_type,
            verbose,
            arena_radius_cm,
            running_window_length,
            window_step_size,
            sample_freq,
            time_bin_size,
            trim,
            all_tracks,
        )
    if file_type == "AnyMaze Head":
        all_tracks = anymaze_tracker.read_anymaze_head(
            groups_with_file_type,
            verbose,
            arena_radius_cm,
            running_window_length,
            window_step_size,
            sample_freq,
            time_bin_size,
            trim,
            all_tracks,
        )

    return all_tracks



import opynfield.readin.track
from opynfield.readin import read_in
import numpy as np


def run_all_track_types(
    groups_and_types: dict[str, list[str]],
    verbose: bool,
    arena_radius_cm: float,
    running_window_length: int,
    window_step_size: int,
    sample_freq: int,
    time_bin_size: int,
    trim: int,
) -> list[opynfield.readin.track.Track]:
    """This function coordinates the entire read-in process

    :param groups_and_types: the input information of which groups were recorded with which filetypes
    :type groups_and_types: dict[str, list[str]]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param arena_radius_cm: the radius of the arena in which the track was recorded, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type arena_radius_cm: float
    :param running_window_length: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type running_window_length: int
    :param window_step_size: a smoothing function parameter set to match ethovision, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type window_step_size: int
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param trim: how many points are recorded before the animal enters the arena (maximum of all animals in this track type), sourced from :class:`opynfield.config.user_input.UserInput` object
    :type trim: int
    :return: all the Tracks that were read in
    :rtype: list[opynfield.readin.track.Track]
    """
    file_types_included = groups_to_types(groups_and_types)  # what file types to run
    print(file_types_included)
    groups_by_filetype = types_to_groups(file_types_included, groups_and_types)
    print(groups_by_filetype)
    all_tracks = list()
    for f in file_types_included:
        print(f"Read In {f} Files For Groups {groups_by_filetype[f]}")
        all_tracks = read_in.read_track_types(
            f,
            groups_by_filetype[f],
            verbose,
            arena_radius_cm,
            running_window_length,
            window_step_size,
            sample_freq,
            time_bin_size,
            trim,
            all_tracks,
        )
    return all_tracks


def groups_to_types(groups_and_types: dict[str, list[str]]) -> list[str]:
    """This function makes a list of which file types are used in the run

    :param groups_and_types: the input information of which groups were recorded with which filetypes
    :type groups_and_types: dict[str, list[str]]
    :return: the filetypes included in the run
    :rtype: list[str]
    """
    # make lists of filetypes in this run
    file_types_included = list()
    for group in groups_and_types:
        for f_type in groups_and_types[group]:
            file_types_included.append(f_type)
    file_types_included = list(np.unique(np.array(file_types_included)))
    return file_types_included


def types_to_groups(
    file_types_included: list[str], groups_and_types: dict[str, list[str]]
) -> dict[str, list[str]]:
    """This function creates a list for each filetype of which groups were recorded in that filetype

    :param file_types_included: which filetypes were included in the run
    :type file_types_included: list[str]
    :param groups_and_types: the input information of which groups were recorded with which filetypes
    :type groups_and_types: dict[str, list[str]]
    :return: which groups were recorded in which filetypes
    :rtype: dict[str, list[str]]
    """
    # for each type in the run, make a list of groups in that type
    groups_by_filetype = dict()  # store results
    for f_type in file_types_included:
        # find all the groups that have this file type
        groups = list()  # list of groups with file type
        for gs in groups_and_types:
            # does this group include this file type
            if f_type in groups_and_types[gs]:
                groups.append(gs)  # if so add it
        groups_by_filetype[f_type] = groups
    return groups_by_filetype



from dataclasses import dataclass
from typing import Union
import numpy as np


@dataclass()
class ExponentialModel:
    """This dataclass groups together relevant information for exponential model fits

    Attributes:
        initial_params (tuple[float, float, float): what p0 to use in curve fitting, defaults to (0.1, -0.1, 0.01)
        bounds (tuple[tuple[float, float, float], tuple[float, float, float]]): what bounds to use in curve fitting, defaults to ((0, -10, 0), (10, 0, 10))
        max_eval (int): max number of iterations the curve fitting process can take, defaults to 4000
        display_parts (tuple[str]): strings that when combined with parameter fit values will show the equation of the best fit model, defaults to ("y = ", " * e ^ (", " * x) + ")

    Methods:
        model_function: the exponential model function
    """
    initial_params: tuple[float, float, float] = (0.1, -0.1, 0.01)
    bounds: tuple[tuple[float, float, float], tuple[float, float, float]] = (
        (0, -10, 0),
        (10, 0, 10),
    )
    max_eval: int = 4000
    display_parts: tuple[str] = ("y = ", " * e ^ (", " * x) + ")

    @staticmethod
    def model_function(x, a, b, c):
        # Exponential Implementation
        y = a * np.exp(b * x) + c
        return y


@dataclass()
class FixedExponentialModel:
    """This dataclass groups together relevant information for fixed exponential model fits

    Attributes:
        initial_params (tuple[float, float): what p0 to use in curve fitting, defaults to (-0.1, -0.1)
        bounds (tuple[tuple[float, float], tuple[float, float]]): what bounds to use in curve fitting, defaults to ((-10, -10), (0, 0))
        max_eval (int): max number of iterations the curve fitting process can take, defaults to 4000
        display_parts (tuple[str]): strings that when combined with parameter fit values will show the equation of the best fit model, defaults to ("y = ", " * (e ^ (", " * x) - 1)")

    Methods:
        model_function: the fixed exponential model function
    """
    initial_params: tuple[float, float] = (-0.1, -0.1)
    bounds: tuple[tuple[float, float], tuple[float, float]] = ((-10, -10), (0, 0))
    max_eval: int = 4000
    display_parts: tuple[str] = ("y = ", " * (e ^ (", " * x) - 1)")

    @staticmethod
    def model_function(x, a, b):
        # Exponential Implementation
        y = a * (np.exp(b * x) - 1)
        return y


@dataclass()
class LinearIncreaseModel:
    """This dataclass groups together relevant information for linear increase model fits

    Attributes:
        initial_params (tuple[float, float): what p0 to use in curve fitting, defaults to (0.1, 0.1)
        bounds (tuple[tuple[float, float], tuple[float, float]]): what bounds to use in curve fitting, defaults to ((0, 0), (10, 10))
        max_eval (int): max number of iterations the curve fitting process can take, defaults to 4000
        display_parts (tuple[str]): strings that when combined with parameter fit values will show the equation of the best fit model, defaults to ("y = ", " * x + ")

    Methods:
        model_function: the linear increase model function
    """
    initial_params: tuple[float, float] = (0.1, 0.1)
    bounds: tuple[tuple[float, float], tuple[float, float]] = ((0, 0), (10, 10))
    max_eval: int = 4000
    display_parts: tuple[str] = ("y = ", " * x + ")

    @staticmethod
    def model_function(x, a, b):
        # Exponential Implementation
        y = (a * x) + b
        return y


@dataclass()
class LinearDecreaseModel:
    """This dataclass groups together relevant information for linear decrease model fits

    Attributes:
        initial_params (tuple[float, float): what p0 to use in curve fitting, defaults to (-0.1, 0.1)
        bounds (tuple[tuple[float, float], tuple[float, float]]): what bounds to use in curve fitting, defaults to ((-10, 0), (0, 10))
        max_eval (int): max number of iterations the curve fitting process can take, defaults to 4000
        display_parts (tuple[str]): strings that when combined with parameter fit values will show the equation of the best fit model, defaults to ("y = ", " * x + ")

    Methods:
        model_function: the linear decrease model function
    """
    initial_params: tuple[float, float] = (-0.1, 0.1)
    bounds: tuple[tuple[float, float], tuple[float, float]] = ((-10, 0), (0, 10))
    max_eval: int = 4000
    display_parts: tuple[str] = ("y = ", " * x + ")

    @staticmethod
    def model_function(x, a, b):
        # Exponential Implementation
        y = (a * x) + b
        return y


@dataclass
class ModelSpecification:
    """This dataclass groups together a set of x and y axes to model the relationship between and the model (with
    associated model information) to use for that relationship

    Attributes:
        axes (tuple[str, str]): the x and y variables to model against each other
        model (Union[ExponentialModel, FixedExponentialModel, LinearIncreaseModel, LinearDecreaseModel]): the model to use to model the axes

    Methods:
        get_x: isolate the x-axis from the axes
        get_y: isolate the y-axis from the axes
    """
    axes: tuple[str, str]
    model: Union[
        ExponentialModel,
        FixedExponentialModel,
        LinearIncreaseModel,
        LinearDecreaseModel,
    ]

    def get_x(self):
        return self.axes[0]

    def get_y(self):
        return self.axes[1]


def mapper(
    y: str,
    map_exponential: tuple[str],
    map_fixed: tuple[str],
    map_linear_increase: tuple[str],
    map_linear_decrease: tuple[str],
) -> Union[
    ExponentialModel, FixedExponentialModel, LinearIncreaseModel, LinearDecreaseModel
]:
    """This function maps a y-axis variable to the model type that should be used in modeling its relationships

    :param y: the y-axis variable to map to a model type
    :type y: str
    :param map_exponential: which y-axis variables should be mapped to an exponential model
    :type map_exponential: tuple[str]
    :param map_fixed: which y-axis variables should be mapped to a fixed exponential model
    :type map_fixed: tuple[str]
    :param map_linear_increase: which y-axis variables should be mapped to a linear increase model
    :type map_linear_increase: tuple[str]
    :param map_linear_decrease: which y-axis variables should be mapped to a linear decrease model
    :type map_linear_decrease: tuple[str]
    :return: the model to use
    :rtype: Union[ExponentialModel, FixedExponentialModel, LinearIncreaseModel, LinearDecreaseModel]
    """
    if y in map_exponential:
        return ExponentialModel()
    elif y in map_fixed:
        return FixedExponentialModel()
    elif y in map_linear_decrease:
        return LinearDecreaseModel()
    elif y in map_linear_increase:
        return LinearIncreaseModel()
    else:
        pass


def set_up_fits(
    x_list: tuple[str] = ("time", "coverage", "pica", "pgca", "percent_coverage"),
    y_time_list: tuple[str] = (
        "activity",
        "coverage",
        "percent_coverage",
        "pica",
        "pgca",
        "p_plus_plus",
        "p_plus_minus",
        "p_plus_zero",
        "p_zero_plus",
        "p_zero_zero",
        "coverage",
        "percent_coverage",
        "pica",
        "pgca",
        "p_plus_plus_given_plus",
        "p_plus_minus_given_plus",
        "p_plus_zero_given_plus",
        "p_zero_plus_given_zero",
        "p_zero_zero_given_zero",
        "p_plus_plus_given_any",
        "p_plus_minus_given_any",
        "p_plus_zero_given_any",
        "p_zero_plus_given_any",
        "p_zero_zero_given_any"
    ),
    y_other_list: tuple[str] = (
        "activity",
        "p_plus_plus",
        "p_plus_minus",
        "p_plus_zero",
        "p_zero_plus",
        "p_zero_zero",
        "p_plus_plus_given_plus",
        "p_plus_minus_given_plus",
        "p_plus_zero_given_plus",
        "p_zero_plus_given_zero",
        "p_zero_zero_given_zero",
        "p_plus_plus_given_any",
        "p_plus_minus_given_any",
        "p_plus_zero_given_any",
        "p_zero_plus_given_any",
        "p_zero_zero_given_any"
    ),
    map_exponential: tuple[str] = (
        "activity",
        "p_plus_plus",
        "p_plus_plus_given_plus",
        "p_plus_plus_given_any",
        "p_zero_plus",
        "p_zero_plus_given_zero",
        "p_zero_plus_given_any"
    ),
    map_fixed: tuple[str] = (
        "p_plus_minus",
        "p_plus_minus_given_plus",
        "p_plus_minus_given_any",
        "p_plus_zero",
        "p_plus_zero_given_plus",
        "p_plus_zero_given_any",
        "p_zero_zero",
        "p_zero_zero_given_zero",
        "p_zero_zero_given_any",
        "coverage",
        "percent_coverage",
        "pica",
        "pgca",
    ),
    map_linear_increase: tuple[str] = (),
    map_linear_decrease: tuple[str] = (),
) -> dict[str, dict[str, ModelSpecification]]:
    """This function

    :param x_list: the x-axis variables, defaults to ("time", "coverage", "pica", "pgca", "percent_coverage")
    :type x_list: tuple[str]
    :param y_time_list: the y-axis variables that can be mapped against time, defaults to ("activity", "coverage", "percent_coverage", "pica", "pgca", "p_plus_plus", "p_plus_minus", "p_plus_zero", "p_zero_plus", "p_zero_zero", "coverage", "percent_coverage", "pica", "pgca", "p_plus_plus_given_plus", "p_plus_minus_given_plus", "p_plus_zero_given_plus", "p_zero_plus_given_zero", "p_zero_zero_given_zero", "p_plus_plus_given_any", "p_plus_minus_given_any", "p_plus_zero_given_any", "p_zero_plus_given_any", "p_zero_zero_given_any")
    :type y_time_list: tuple[str]
    :param y_other_list: the y-axis variables that can be mapped against coverage measures, defaults to ("activity", "p_plus_plus", "p_plus_minus", "p_plus_zero", "p_zero_plus", "p_zero_zero", "p_plus_plus_given_plus", "p_plus_minus_given_plus", "p_plus_zero_given_plus", "p_zero_plus_given_zero", "p_zero_zero_given_zero", "p_plus_plus_given_any", "p_plus_minus_given_any", "p_plus_zero_given_any", "p_zero_plus_given_any", "p_zero_zero_given_any")
    :type y_other_list: tuple[str]
    :param map_exponential: the y-axis variables that should be modeled with an exponential model, defaults to ("activity", "p_plus_plus", "p_plus_plus_given_plus", "p_plus_plus_given_any", "p_zero_plus", "p_zero_plus_given_zero", "p_zero_plus_given_any",)
    :type map_exponential:  tuple[str]
    :param map_fixed: the y-axis variables that should be modeled with a fixed exponential model, defaults to ("p_plus_minus", "p_plus_minus_given_plus", "p_plus_minus_given_any", "p_plus_zero", "p_plus_zero_given_plus", "p_plus_zero_given_any", "p_zero_zero", "p_zero_zero_given_zero", "p_zero_zero_given_any", "coverage", "percent_coverage", "pica", "pgca",)
    :type map_fixed: tuple[str]
    :param map_linear_increase: the y-axis variables that should be modeled with a linear increase model, defaults to ()
    :type map_linear_increase: tuple[str]
    :param map_linear_decrease: the y-axis variables that should be modeled with a linear decrease model, defaults to ()
    :type map_linear_decrease: tuple[str]
    :return: a dictionary of x-axis to a dictionary of y-axis to a model type to use
    :rtype: dict[str, dict[str, ModelSpecification]]
    """
    xy_dict = {}
    for x in x_list:
        x_dict = {}
        if x == "time":
            for y in y_time_list:
                x_dict[y] = ModelSpecification(
                    (x, y),
                    mapper(
                        y,
                        map_exponential,
                        map_fixed,
                        map_linear_increase,
                        map_linear_decrease,
                    ),
                )
        else:
            for y in y_other_list:
                x_dict[y] = ModelSpecification(
                    (x, y),
                    mapper(
                        y,
                        map_exponential,
                        map_fixed,
                        map_linear_increase,
                        map_linear_decrease,
                    ),
                )
        xy_dict[x] = x_dict
    return xy_dict


# TODO: Change the method of creating the xy dict so that it is easier for the user to pass in inputs
# keep it so that the xy dict is indexed first by x then by y to get a ModelSpecification



from tkinter import filedialog
import openpyxl as xl
import pandas as pd
from opynfield.readin.track import Track
from opynfield.readin import multi_tracker
import numpy as np
import xlrd


def read_etho_v1(
    groups_with_file_type: list[str],
    verbose: bool,
    sample_freq: int,
    time_bin_size: int,
    all_tracks: list[Track],
) -> list[Track]:
    """This function reads in all the tracks from the Ethovision Version 1 tracker format. The function extracts the x,
    y, and t information and smooths, centers, and converts the units of the track.

    :param groups_with_file_type: which groups have tracks recorded in this type
    :type groups_with_file_type: list[str]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param all_tracks: a list with all the Track objects from previously read-in datatypes
    :type all_tracks: list[Track]
    :return: a Track object with a consistent format for x y and t tracking points
    :rtype: list[Track]
    """
    for etho_group in groups_with_file_type:
        if verbose:
            print(f"Running Ethovision V1 Tracker Files For Group: {etho_group}")
    title1 = "Select all .xls v1 files for this experiment"
    # select excel files for all groups
    # noinspection PyArgumentList
    all_group_data_files = filedialog.askopenfilenames(
        filetypes=[("Excel Files", "*.xls *.xlsx")], title=title1, multiple=True
    )
    etho_tracks = list()
    for file_num in range(len(all_group_data_files)):
        file = all_group_data_files[file_num]
        if verbose:
            print(f"Running File {file_num + 1} Out Of {len(all_group_data_files)}")
        wb = xl.load_workbook(file)
        # read in each file and sheet's data
        for sheet in wb:
            header_lines = int(sheet["B1"].value)
            # track info
            group = sheet["B35"].value
            arena = sheet["B6"].value
            parts = arena.split()
            sheet_num = int(parts[1])  # sheet num indicated which arena it came from
            if verbose:
                print(f"Running Sheet {sheet_num}")
            # track data
            df = pd.DataFrame(sheet.values)
            time_col = df[1][header_lines:]
            x_col = df[2][header_lines:]
            y_col = df[3][header_lines:]
            if y_col[38] is None:
                print(f"No data in file {file_num + 1} sheet {sheet_num}")
            else:
                track = Track(
                    group,
                    x_col.values,
                    y_col.values,
                    time_col.values,
                    "Ethovision Excel Version 1",
                    [sheet_num],
                    False,
                )
                # put track coordinates through standardization procedures (for single track)
                track.etho_v1_numeric(verbose)
                track.etho_v1_subsample(sample_freq, time_bin_size, verbose)
                track.etho_v1_fill_missing(verbose)
                etho_tracks.append(track)
    # group tracks by arena and get their center points
    tracks_by_arena = sort_tracks_by_arena(etho_tracks)
    combined_coords_by_arena = combine_arena_coords(tracks_by_arena)
    center_points_by_area = extract_arena_center_point(
        combined_coords_by_arena, verbose
    )
    for track in etho_tracks:
        # put track coordinates through standardization procedures (given the arena)
        track.etho_v1_convert_to_center(center_points_by_area, verbose)
        track.standardized = True
        all_tracks.append(track)
    return all_tracks


def read_etho_v2(
    groups_with_file_type: list[str],
    verbose: bool,
    sample_freq: int,
    time_bin_size: int,
    all_tracks: list[Track],
) -> list[Track]:
    """This function reads in all the tracks from the Ethovision Version 2 tracker format. The function extracts the x,
    y, and t information and smooths, centers, and converts the units of the track.

    :param groups_with_file_type: which groups have tracks recorded in this type
    :type groups_with_file_type: list[str]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param all_tracks: a list with all the Track objects from previously read-in datatypes
    :type all_tracks: list[Track]
    :return: a Track object with a consistent format for x y and t tracking points
    :rtype: list[Track]
    """
    for etho_group in groups_with_file_type:
        if verbose:
            print(f"Running Ethovision V2 Tracker Files For Group: {etho_group}")
    title1 = "Select all .xls v2 files for this experiment"
    # select excel files for all groups
    # noinspection PyArgumentList
    all_group_data_files = filedialog.askopenfilenames(
        filetypes=[("Excel Files", "*.xls *.xlsx")], title=title1, multiple=True
    )
    etho_tracks = list()
    for file_num in range(len(all_group_data_files)):
        file = all_group_data_files[file_num]
        if verbose:
            print(f"Running File {file_num + 1} Out Of {len(all_group_data_files)}")
        wb = xlrd.open_workbook(file)
        # read in each file and sheet's data
        for sheet in wb:
            header_lines = 28
            print("Warning, make sure this file has 28 header lines")
            # track info
            group = sheet[25, 1].value
            arena = sheet[10, 1].value
            parts = arena.split()
            sheet_num = int(parts[1])  # sheet num indicates which arena it came from
            if verbose:
                print(f"Running Sheet {sheet_num}")
            # track data
            x = list()
            y = list()
            t = list()
            for i in range(sheet.nrows):
                t.append(sheet[i, 0].value)
                x.append(sheet[i, 1].value)
                y.append(sheet[i, 2].value)
            time_col = np.array(t[header_lines:sheet.nrows])
            x_col = np.array(x[header_lines:sheet.nrows])
            y_col = np.array(y[header_lines:sheet.nrows])
            track = Track(
                group,
                x_col,
                y_col,
                time_col,
                "Ethovision Excel Version 2",
                [sheet_num],
                False,
            )
            # put track coordinates through standardization procedures (for single track)
            track.etho_v2_numeric(verbose)
            track.etho_v2_subsample(sample_freq, time_bin_size, verbose)
            track.etho_v2_fill_missing(verbose)
            etho_tracks.append(track)
    # group tracks by arena and get their center points
    tracks_by_arena = sort_tracks_by_arena(etho_tracks)
    combined_coords_by_arena = combine_arena_coords(tracks_by_arena)
    center_points_by_area = extract_arena_center_point(
        combined_coords_by_arena, verbose
    )
    for track in etho_tracks:
        # put track coordinates through standardization procedures (given the arena)
        track.etho_v2_convert_to_center(center_points_by_area, verbose)
        track.standardized = True
        all_tracks.append(track)
    return all_tracks


def read_etho_txt(
    groups_with_file_type: list[str],
    verbose: bool,
    sample_freq: int,
    time_bin_size: int,
    all_tracks: list[Track],
) -> list[Track]:
    """This function reads in all the tracks from the Ethovision Text Version tracker format. The function extracts the
    x, y, and t information and smooths, centers, and converts the units of the track.

    :param groups_with_file_type: which groups have tracks recorded in this type
    :type groups_with_file_type: list[str]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param all_tracks: a list with all the Track objects from previously read-in datatypes
    :type all_tracks: list[Track]
    :return: a Track object with a consistent format for x y and t tracking points
    :rtype: list[Track]
    """
    for etho_group in groups_with_file_type:
        if verbose:
            print(f"Running Ethovision Text Tracker Files For Group: {etho_group}")
    title1 = "Select all .txt files for this experiment"
    # select text files for all groups
    # noinspection PyArgumentList
    all_group_data_files = filedialog.askopenfilenames(
        filetypes=[("Text Files", "*.txt")], title=title1, multiple=True
    )
    etho_tracks = list()
    # need to force column names since data is saved in txt format
    labels = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K"]

    # read in each file's data
    for file_num in range(len(all_group_data_files)):
        file = all_group_data_files[file_num]
        if verbose:
            print(f"Running File {file_num + 1} Out Of {len(all_group_data_files)}")
        df = pd.read_csv(file, encoding="UTF-16", names=labels, sep=",")
        # track info
        header_lines = df["B"][0]  # should be 38
        arena = df["B"][5]
        parts = arena.split()
        sheet_num = int(parts[1])  # sheet num indicates which arena it came from
        group = df["B"][34]
        # track data
        t = df["B"][header_lines:].values
        x = df["C"][header_lines:].values
        y = df["D"][header_lines:].values
        # convert to float
        for c in range(len(t)):
            try:
                t[c] = float(t[c])
                x[c] = float(x[c])
                y[c] = float(y[c])
            except ValueError:
                pass  # nans and blank lines at end of file
        track = Track(group, x, y, t, "Ethovision Text", [sheet_num], False)
        # put track coordinates through standardization procedures (for single track)
        track.etho_txt_numeric(verbose)
        track.etho_txt_subsample(sample_freq, time_bin_size, verbose)
        track.etho_txt_fill_missing(verbose)
        etho_tracks.append(track)
    # group tracks by arena and get their center points
    tracks_by_arena = sort_tracks_by_arena(etho_tracks)
    combined_coords_by_arena = combine_arena_coords(tracks_by_arena)
    center_points_by_area = extract_arena_center_point(
        combined_coords_by_arena, verbose
    )
    for track in etho_tracks:
        # put track coordinates through standardization procedures (given the arena)
        track.etho_txt_convert_to_center(center_points_by_area, verbose)
        track.standardized = True
        all_tracks.append(track)
    return all_tracks


def read_etho_ml(
    groups_with_file_type: list[str],
    verbose: bool,
    sample_freq: int,
    time_bin_size: int,
    all_tracks: list[Track],
) -> list[Track]:
    """This function reads in all the tracks from the Ethovision TextML Version tracker format. The function extracts
    the x, y, and t information and smooths, centers, and converts the units of the track.

    :param groups_with_file_type: which groups have tracks recorded in this type
    :type groups_with_file_type: list[str]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :param sample_freq: the frame rate that the track was recorded with, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type sample_freq: int
    :param time_bin_size: how many seconds should be aggregated together, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type time_bin_size: int
    :param all_tracks: a list with all the Track objects from previously read-in datatypes
    :type all_tracks: list[Track]
    :return: a Track object with a consistent format for x y and t tracking points
    :rtype: list[Track]
    """
    for etho_group in groups_with_file_type:
        if verbose:
            print(f"Running Ethovision ML Tracker Files For Group: {etho_group}")
        title1 = f"Select ML .txt files in folder for {etho_group}"
        # select ML .txt files from ethovision
        data_files = filedialog.askopenfilenames(
            filetypes=[("Text Files", "*.txt")], title=title1
        )
        for file_num in range(len(data_files)):
            if verbose:
                print(f"{etho_group}, File{file_num + 1} Out Of {len(data_files)}")
            # this file type lacks the track info, only contains the track data
            file_data = pd.read_csv(data_files[file_num], sep="\t", header=None)
            track = Track(
                etho_group,
                file_data[1].values,
                file_data[2].values,
                file_data[0].values,
                "Ethovision Through MATLAB",
                [],
                False,
            )
            # put track coordinates through standardization procedures (for single track)
            track.etho_ml_numeric(verbose)
            track.etho_ml_subsample(sample_freq, time_bin_size, verbose)
            track.etho_ml_fill_missing(verbose)
            # since we don't know which arena the track came from, we must calculate the center point
            # from the track's coordinates, rather than combined coordinates
            track_center = multi_tracker.calc_center(track.x, track.y, verbose)
            track.etho_ml_convert_to_center(track_center, verbose)
            track.standardized = True
            all_tracks.append(track)
    return all_tracks


def sort_tracks_by_arena(list_of_etho_tracks: list[Track]) -> dict[str, list[Track]]:
    """This function creates a dictionary of Track objects indexed by which ethovision arena they were recorded in

    :param list_of_etho_tracks: the Tracks to sort by arena
    :type list_of_etho_tracks: list[Track]
    :return: the dictionary of tracks by arena
    :rtype: dict[str, list[Track]]
    """
    tracks_by_arena = dict()
    for track in list_of_etho_tracks:
        if track.options[0] not in tracks_by_arena:
            # if there hasn't been a track from that arena, add the arena to the dict with that track
            tracks_by_arena[track.options[0]] = [track]
        else:
            # if there are already tracks from that arena, add the track to the list of tracks from that arena
            tracks_by_arena[track.options[0]].append(track)
    return tracks_by_arena


def combine_arena_coords(
    tracks_by_arena: dict[str, list[Track]],
) -> dict[str, tuple[np.ndarray, np.ndarray]]:
    """This function combines the coordinates of all the tracks that were recorded in the same arena so that the arena
    center point can be estimated

    :param tracks_by_arena: the dict of tracks recorded in each arena
    :type tracks_by_arena: dict[str, list[Track]]
    :return: the combined x and y coordinates, indexed by arena
    :rtype: dict[str, tuple[np.ndarray, np.ndarray]]
    """
    combined_coords_by_arena = dict()
    for arena in tracks_by_arena:
        # the first track initializes the combined coordinates
        combined_x = tracks_by_arena[arena][0].x
        combined_y = tracks_by_arena[arena][0].y
        for track_num in range(1, len(tracks_by_arena[arena])):
            # for the rest of the tracks
            combined_x = np.append(combined_x, tracks_by_arena[arena][track_num].x)
            combined_y = np.append(combined_y, tracks_by_arena[arena][track_num].y)
        # once all tracks run in that arena are added, save them to the dict
        combined_coords_by_arena[arena] = (combined_x, combined_y)
    return combined_coords_by_arena


def extract_arena_center_point(
    combined_coords_by_arena: dict[str, tuple[np.ndarray, np.ndarray]], verbose: bool
) -> dict[str, tuple[float, float]]:
    """This function estimates the center point of the arena that tracks were recorded in

    :param combined_coords_by_arena: the combined x and y coordinates, indexed by arena
    :type combined_coords_by_arena: dict[str, tuple[np.ndarray, np.ndarray]]
    :param verbose: display progress update, sourced from :class:`opynfield.config.user_input.UserInput` object
    :type verbose: bool
    :return: the center points of the arenas, indexed by arena name
    :rtype: dict[str, tuple[float, float]]
    """
    center_points_by_area = dict()
    for arena in combined_coords_by_arena:
        # for each arena, calculate the center point from the combined coordinates of all tracks run in that arena
        center_point = multi_tracker.calc_center(
            combined_coords_by_arena[arena][0],
            combined_coords_by_arena[arena][1],
            verbose,
        )
        if verbose:
            print(f"Arena {arena} Center Point: {center_point}")
        center_points_by_area[arena] = center_point
    return center_points_by_area



import pandas as pd
from opynfield.config.defaults_settings import Defaults
from opynfield.config.user_input import UserInput
import os
from copy import deepcopy
from statsmodels.multivariate.manova import MANOVA
from statsmodels.formula.api import ols
from dataclasses import dataclass
from statsmodels.regression.linear_model import RegressionResultsWrapper
from statsmodels.stats.anova import anova_lm
import sys


def format_params(
    bounded_fits: dict[str, dict[str, dict[str, pd.DataFrame]]],
    defaults: Defaults,
    user_inputs: UserInput,
) -> pd.DataFrame:
    """This function takes the parameter fits of the individuals bounded models and saves them out to .csv files

    :param bounded_fits: the parameters to re-format
    :type bounded_fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param user_inputs: the user inputs to use
    :type user_inputs: UserInput
    :return: the formatted individual parameters
    :rtype: pd.DataFrame
    """
    dict_groups = {}
    for group in bounded_fits:
        group_df_list = []
        for x_measure in bounded_fits[group]:
            for y_measure in bounded_fits[group][x_measure]:
                df = bounded_fits[group][x_measure][y_measure]
                df = df.rename(
                    columns=lambda x: f"{x_measure}_{y_measure}_parameter_{x}"
                )
                group_df_list.append(df)
        full_group = pd.concat(group_df_list, axis=1)
        dict_groups[group] = full_group
        if defaults.save_group_model_csvs:
            # make a dir for results -> individuals -> group -> models
            path = (
                user_inputs.result_path
                + "/Individuals/"
                + user_inputs.groups_to_paths[group]
                + "/Models"
            )
            os.makedirs(path, exist_ok=True)
            df_path = path + "/IndividualModels.csv"
            full_group.to_csv(path_or_buf=df_path, index=False)
    for group, df in dict_groups.items():
        df["group"] = group
    f_df = pd.concat(dict_groups.values(), ignore_index=True)
    if defaults.save_all_group_model_csvs:
        path = user_inputs.result_path + "/Individuals/CombinedGroups/Models"
        os.makedirs(path, exist_ok=True)
        df_path = path + "/CombinedGroups_IndividualModels.csv"
        f_df.to_csv(path_or_buf=df_path, index=False)
    return f_df


def format_group_params(
    group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]],
    defaults: Defaults,
    user_inputs: UserInput,
):
    """This function takes the parameter fits of the group models and saves them out to .csv files

    :param group_fits: the parameters to re-format
    :type group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param user_inputs: the user inputs to use
    :type user_inputs: UserInput
    """
    dict_groups = {}
    for group in group_fits:
        group_df_list = []
        for x_measure in group_fits[group]:
            for y_measure in group_fits[group][x_measure]:
                df = group_fits[group][x_measure][y_measure].T
                df = df.rename(
                    columns=lambda x: f"{x_measure}_{y_measure}_parameter_{x}"
                )
                group_df_list.append(df)
        full_group = pd.concat(group_df_list, axis=1)
        dict_groups[group] = full_group
        if defaults.save_group_model_csvs:
            # make a dir for results -> groups -> group -> models
            path = (
                user_inputs.result_path
                + "/Groups/"
                + user_inputs.groups_to_paths[group]
                + "/Models"
            )
            os.makedirs(path, exist_ok=True)
            df_path = path + "/GroupModels.csv"
            full_group.to_csv(path_or_buf=df_path, index=False)
    for group, df in dict_groups.items():
        df["group"] = group
    f_df = pd.concat(dict_groups.values(), ignore_index=True)
    if defaults.save_all_group_model_csvs:
        path = user_inputs.result_path + "/Groups/CombinedGroups/Models"
        os.makedirs(path, exist_ok=True)
        df_path = path + "/CombinedGroups_GroupModels.csv"
        f_df.to_csv(path_or_buf=df_path, index=False)
    return


def create_full_formula(test_df):
    """This function creates the formula for the full MANOVA test (of the form 'group ~ param1 + param 2 + ...')

    :param test_df: df of the parameter fits for each individual
    :type test_df: pd.DataFrame
    :return: the full formula
    :rtype: str
    """
    vals_to_test = list(test_df.columns)
    vals_to_test.remove("group")
    formula = ""
    for v in range(len(vals_to_test)):
        if v < len(vals_to_test) - 1:
            formula = formula + vals_to_test[v] + " + "
        else:
            formula = formula + vals_to_test[v]
    formula = formula + " ~ group"
    return formula


def run_full_test(formatted_bounded_fits: pd.DataFrame):
    """This function runs the MANOVA test on every parameter by group

    :param formatted_bounded_fits: df of the parameter fits for each individual
    :type formatted_bounded_fits: pd.DataFrame
    :return: a deepcopy of the df of parameters (needed to memory consolidation)
    :rtype: pd.DataFrame
    """
    test_df = deepcopy(formatted_bounded_fits).copy()
    full_formula = create_full_formula(test_df)
    model_full = MANOVA.from_formula(full_formula, data=test_df)
    print("Full Model MANOVA Results")
    print("\n")
    print(model_full.mv_test())
    return test_df


def get_col_names_from_prefix(test_df: pd.DataFrame, prefix: str) -> list[str]:
    """This function generates a list of parameters that have the same prefix as the parameter being tested (the
    parameters that belong to the same model fit)

    :param test_df: df of the parameter fits for each individual
    :type test_df: pd.DataFrame
    :param prefix: the parameter name beginning
    :type prefix: str
    :return: the parameters from the same model
    :rtype: list[str]
    """
    cols_that_match = list()
    for c in test_df:
        if c.startswith(prefix):
            cols_that_match.append(c)
    return cols_that_match


def make_formula_from_list(dependent_vars: list[str], independent_var: str) -> str:
    """This function creates the formula for the MANOVA of all the parameters from the same model fit

    :param dependent_vars: the list of parameters from the same model fit
    :type dependent_vars: list[str]
    :param independent_var: the independent variable of the formula
    :type independent_var: str
    :return: the formula for the test
    :rtype: str
    """
    formula = " + ".join(dependent_vars) + f" ~ {independent_var}"
    return formula


@dataclass
class ManovaModels:
    """This dataclass associates information for the MANOVA, ANOVA, and t-test models of a specific parameter set

    Attributes:
        full_model (MANOVA): the full MANOVA model for all the parameters in the model parameter set
        component_model_results (dict[str, RegressionResultsWrapper]): a dict of the ANOVA model for each parameter in the set
        independent_var (str): the independent variable in the models

    Methods:
        parameter_names: supplies the list of parameters in the sets
        display_results: prints results of MANOVA, ANOVA, and t-test models
    """
    full_model: MANOVA
    component_model_results: dict[str, RegressionResultsWrapper]
    independent_var: str

    def parameter_names(self):
        """This method returns all the parameters used in the MANOVA models

        :return: the parameters in the set
        :rtype: list[str]
        """
        return list(self.component_model_results.keys())

    def display_results(self):
        """This method prints the results of all the MANOVA, ANOVA, and t-test models"""
        p_names = self.parameter_names()
        name = "_".join(p_names[0].split("_")[:-1]) + "s"

        # display the full model results
        print(f"{name} MANOVA Results")
        print("\n")
        print(self.full_model.mv_test())

        # display the parameter model ANOVA results
        # and the pairwise t test
        for p in self.component_model_results:
            print(f"{p} ANOVA Results")
            print("\n")
            print((anova_lm(self.component_model_results[p])))
            print("\n")
            print(f"{p} Pairwise T-test Results")
            print("\n")
            frame = (
                self.component_model_results[p]
                .t_test_pairwise(self.independent_var)
                .result_frame
            )
            print(frame)
            print("\n")
        return


def generate_models(test_df: pd.DataFrame, prefix: str, independent_var: str = "group"):
    """This function generates the ANOVA and t-test models for a specific parameter

    :param test_df: df of the parameter fits for each individual
    :type test_df: pd.DataFrame
    :param prefix: the parameter name
    :type prefix: str
    :param independent_var: the independent variable of the test, defaults to group
    :type independent_var: str
    :return: a ManovaModels instance containing the test information for the parameter
    :rtype: ManovaModels
    """
    cols_that_match = get_col_names_from_prefix(test_df, prefix)
    full_formula = make_formula_from_list(
        dependent_vars=cols_that_match, independent_var=independent_var
    )
    # full model
    model_f = MANOVA.from_formula(full_formula, data=test_df)
    # model for each var
    component_model_results = dict()
    for c in cols_that_match:
        formula_c = make_formula_from_list(
            dependent_vars=[c], independent_var=independent_var
        )
        model_result_c = ols(formula_c, test_df).fit()
        component_model_results[c] = model_result_c
    models = ManovaModels(
        full_model=model_f,
        component_model_results=component_model_results,
        independent_var=independent_var,
    )
    return models


def run_component_tests(test_df: pd.DataFrame, defaults: Defaults):
    """This function coordinates running follow-up ANOVA and t-tests for each parameter from the full model

    :param test_df: df of the parameter fits for each individual
    :type test_df: pd.DataFrame
    :param defaults: the default settings to use
    :type defaults: Defaults
    """
    for test_param in defaults.create_pairs():
        models = generate_models(test_df, test_param)
        models.display_results()
    return


def run_tests(
    formatted_bounded_fits: pd.DataFrame, defaults: Defaults, user_inputs: UserInput
):
    """This function coordinates running all the relevant statistical tests on the individual parameters

    :param formatted_bounded_fits: df of the parameter fits for each individual
    :type formatted_bounded_fits: pd.DataFrame
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param user_inputs: the user inputs to use
    :type user_inputs: UserInput
    """
    os.makedirs(user_inputs.result_path + "/Stats")
    stats_results_file = user_inputs.result_path + "/Stats/results.txt"
    orig_stdout = sys.stdout
    f = open(stats_results_file, "a")
    sys.stdout = f
    test_df = run_full_test(formatted_bounded_fits)
    run_component_tests(test_df, defaults)
    sys.stdout = orig_stdout
    f.close()
    return



import pandas as pd
from opynfield.config.model_settings import ModelSpecification
from opynfield.config.defaults_settings import Defaults
from opynfield.config.plot_settings import PlotSettings
from opynfield.config.user_input import UserInput
import matplotlib.pyplot as plt
import numpy as np
from itertools import chain, zip_longest
import os


def truncate(n, decimals=0):
    """This function trims the exact parameter fit value to a set number of decimal places for display on a plot

    :param n: the parameter
    :type n: float
    :param decimals: the number of decimal places to trim to
    :type decimals: int
    :return: the trimmed parameter value
    :rtype: float
    """
    multiplier = 10**decimals
    return int(n * multiplier) / multiplier


def generate_group_equation_str(fit_params, display_parts):
    """This function generates the equation of the best fit model from the model display parts and the truncated
    parameter values

    :param fit_params: the parameter fits
    :type fit_params: list[float]
    :param display_parts: the display parts from the model information
    :type display_parts: list[str]
    :return: the full equation string
    :rtype: str
    """
    equation_parts = list(display_parts)
    string_params = [str(truncate(x, 4)) for x in fit_params]
    parts = [
        x for x in chain(*zip_longest(equation_parts, string_params)) if x is not None
    ]
    display_equation = "".join(parts)
    return display_equation


def generate_fig_title(
    path: str,
    x_measure: str,
    y_measure: str,
    model_fit: bool,
    error: bool,
    extension: str,
):
    """This function generates a figure title to save a single group plot at, based on the component parts

    :param path: the path to the results plot folder
    :type path: str
    :param x_measure: the x-measure of the plot
    :type x_measure: str
    :param y_measure: the y-measure of the plot
    :type y_measure: str
    :param model_fit: whether the plot includes the model fit
    :type model_fit: bool
    :param error: whether the plot includes the group average error bars
    :type error: bool
    :param extension: the file extension / format to save the plot in
    :return: the path to save the plot at
    :rtype: str
    """
    path = path + f"{x_measure}_vs_{y_measure}"
    if model_fit:
        path = path + "_with_model"
    if model_fit and error:
        path = path + "_and"
    if error:
        path = path + "_with_error"
    path = path + extension
    return path


def plot_solo_group_measure_by_time(
    measure_by_time: pd.DataFrame,
    plot_settings: PlotSettings,
    model_specs: ModelSpecification,
    group_fits: pd.DataFrame,
    measure: str,
    group: str,
    user_config: UserInput,
):
    """This function plots a single group's average time vs y-measure relationship

    :param measure_by_time: the y-measure data (and error) for the group
    :type measure_by_time: pd.DataFrame
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param model_specs: the model settings to use
    :type model_specs: ModelSpecification
    :param group_fits: the y-fit data from the model fits
    :type group_fits: pd.DataFrame
    :param measure: the y-measure to plot
    :type measure: str
    :param group: the whose data you are plotting
    :type group: str
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    # create figure and axes objects for the plot
    fig, ax = plt.subplots()
    # define the x and y to be plotted
    # y is the measure for the group
    y_plot = measure_by_time.iloc[0][2:]
    # also have the error on that measure
    y_error = measure_by_time.iloc[1][2:]
    # x is time in sec
    x_plot = np.arange(len(y_plot))
    # do error first so that the markers will go on top of them
    if plot_settings.group_error_bars:
        ax.errorbar(
            x_plot,
            y_plot,
            yerr=y_error,
            xerr=None,
            fmt="none",
            ecolor=plot_settings.error_color,
            errorevery=plot_settings.n_between_error,
            elinewidth=plot_settings.error_width,
            alpha=plot_settings.alpha,
        )
    # scatter plot
    ax.scatter(
        x_plot, y_plot, s=plot_settings.marker_size, c=plot_settings.marker_color
    )

    if plot_settings.group_model_fit:
        y_fit = model_specs.model.model_function(x_plot, *group_fits.values)
        ax.plot(x_plot, y_fit, c=plot_settings.fit_color, alpha=plot_settings.alpha)
        if plot_settings.equation:
            equation_str = generate_group_equation_str(
                group_fits.values.flatten().tolist(), model_specs.model.display_parts
            )
            ax.set_title(f"model fit: {equation_str}")
    # add the axis labels
    ax.set_xlabel("time (s)")
    ax.set_ylabel(f"{measure}")
    # add the plot title
    fig.suptitle(f"{measure} by time group {group}")
    # set axis limits
    if measure not in ["activity", "percent_coverage", "pica", "pgca", "coverage"]:
        # ax.set_xlim()
        ax.set_ylim((-0.1, 1.1))
    if plot_settings.display_solo_group_figures:
        # show the figure
        fig.show()
    if plot_settings.save_solo_group_figures:
        # save the figure
        path = (
            user_config.result_path
            + "/Groups/"
            + user_config.groups_to_paths[group]
            + "/Plots/by_time/"
        )
        os.makedirs(path, exist_ok=True)
        fig_path = generate_fig_title(
            path,
            "time",
            measure,
            plot_settings.group_model_fit,
            plot_settings.group_error_bars,
            plot_settings.fig_extension,
        )
        fig.savefig(fname=fig_path, bbox_inches="tight")
    plt.close(fig=fig)
    return


def plot_solo_group_by_time(
    by_time: dict[str, pd.DataFrame],
    defaults: Defaults,
    plot_settings: PlotSettings,
    model_params: dict[str, ModelSpecification],
    group_fits: dict[str, pd.DataFrame],
    group: str,
    user_config: UserInput,
):
    """This function coordinates the plotting of all time vs y-measure relationships for a single group

    :param by_time: the y-measure data (and error) of the group, indexed by y-measure
    :type by_time: dict[str, pd.DataFrame]
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param model_params: the model settings to use, indexed by y-measure
    :type model_params: dict[str, ModelSpecification]
    :param group_fits: the y-fit data from the model fit parameters, indexed by y-measure
    :type group_fits: dict[str, pd.DataFrame]
    :param group: the group to plot
    :param group: str
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    for measure in defaults.time_averaged_measures:
        if measure != "r":
            plot_solo_group_measure_by_time(
                by_time[measure].reset_index().drop(columns="index"),
                plot_settings,
                model_params[measure],
                group_fits[measure],
                measure,
                group,
                user_config,
            )
    return


def plot_solo_group_by_cmeasure(
    by_coverage: pd.DataFrame,
    defaults: Defaults,
    cmeasure: str,
    plot_settings: PlotSettings,
    model_params: dict[str, ModelSpecification],
    group_fits: dict[str, pd.DataFrame],
    group: str,
    user_config: UserInput,
):
    """This function plots a single group's average coverage-measure vs y-measure relationship for each coverage
    averaged y-measure

    :param by_coverage: the x-measure data (and error) and the y-measure data (and error) for the group
    :type by_coverage: pd.DataFrame
    :param defaults: the default settings to use
    :type defaults: Defaults
    :param cmeasure: the coverage measure being used as x-axis
    :type cmeasure: str
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param model_params: the model settings to use, indexed by y-measure
    :type model_params: dict[str, ModelSpecification]
    :param group_fits: the y-fit data from the model fits, indexed by y-measure
    :type group_fits: dict[str, pd.DataFrame]
    :param group: the whose data you are plotting
    :type group: str
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    for measure in defaults.coverage_averaged_measures:
        # create figure and axes objects for the plot
        fig, ax = plt.subplots()
        # define the x and y to be plotted
        # y is the measure for the group
        y_plot = by_coverage[f"{measure} mean"].values
        # also have the error on that measure
        y_error = by_coverage[f"{measure} sem"].values
        # x is the coverage measure for the group
        x_plot = by_coverage[f"{cmeasure} mean"].values
        # also have the error on the cmeasure
        x_error = by_coverage[f"{cmeasure} sem"].values
        # do error first so that the markers will go on top of them
        if plot_settings.group_error_bars:
            ax.errorbar(
                x_plot,
                y_plot,
                yerr=y_error,
                xerr=x_error,
                fmt="none",
                ecolor=plot_settings.error_color,
                errorevery=plot_settings.n_between_error,
                elinewidth=plot_settings.error_width,
                alpha=plot_settings.alpha,
            )
        # scatter plot
        ax.scatter(
            x_plot, y_plot, s=plot_settings.marker_size, c=plot_settings.marker_color
        )
        if plot_settings.group_model_fit:
            model_specs = model_params[measure]
            measure_fits = group_fits[measure]
            y_fit = model_specs.model.model_function(x_plot, *measure_fits.values)
            ax.plot(x_plot, y_fit, c=plot_settings.fit_color, alpha=plot_settings.alpha)
            if plot_settings.equation:
                equation_str = generate_group_equation_str(
                    measure_fits.values.flatten().tolist(),
                    model_specs.model.display_parts,
                )
                ax.set_title(f"model fit: {equation_str}")
        # add the axis labels
        ax.set_xlabel(f"{cmeasure}")
        ax.set_ylabel(f"{measure}")
        # add the plot title
        fig.suptitle(f"{measure} by {cmeasure} group {group}")
        # set axis limits
        if measure != "activity":
            # ax.set_xlim()
            ax.set_ylim((-0.1, 1.1))
        if plot_settings.display_solo_group_figures:
            # show the figure
            fig.show()
        if plot_settings.save_solo_group_figures:
            # save the figure
            path = (
                user_config.result_path
                + "/Groups/"
                + user_config.groups_to_paths[group]
                + "/Plots/by_"
                + cmeasure
                + "/"
            )
            os.makedirs(path, exist_ok=True)
            fig_path = generate_fig_title(
                path,
                cmeasure,
                measure,
                plot_settings.group_model_fit,
                plot_settings.group_error_bars,
                plot_settings.fig_extension,
            )
            fig.savefig(fname=fig_path, bbox_inches="tight")
        plt.close(fig=fig)
    return


def plot_all_solo_groups(
    group_averages: dict[str, dict],
    group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]],
    model_params: dict[str, dict[str, ModelSpecification]],
    test_defaults: Defaults,
    plot_settings: PlotSettings,
    user_config: UserInput,
):
    """This function coordinates the plotting of all the x-measure vs y-measure relationships for each group

    :param group_averages: the group averages to plot, indexed by x-measure, then group (then y-measure)
    :type group_averages: dict[str, dict]
    :param group_fits: the y-fit values (based on the model fit parameters) indexed by group, time, and y-measure
    :type group_fits: dict[str, dict[str, dict[str, pd.DataFrame]]]
    :param model_params: the model settings to use, indexed by x-measure and y-measure
    :type model_params: dict[str, dict[str, ModelSpecification]]
    :param test_defaults: the default settings to use
    :type test_defaults: Defaults
    :param plot_settings: the plot settings to use
    :type plot_settings: PlotSettings
    :param user_config: the user inputs to use
    :type user_config: UserInput
    """
    for group in group_averages["time"]:
        print(f"Plotting Group {group} by time")
        plot_solo_group_by_time(
            group_averages["time"][group],
            test_defaults,
            plot_settings,
            model_params["time"],
            group_fits[group]["time"],
            group,
            user_config,
        )
        for x_measure in ["coverage", "pica", "pgca", "percent_coverage"]:
            print(f"Plotting Group {group} by {x_measure}")
            plot_solo_group_by_cmeasure(
                group_averages[x_measure][group],
                test_defaults,
                x_measure,
                plot_settings,
                model_params[x_measure],
                group_fits[group][x_measure],
                group,
                user_config,
            )
    return
